"""Regge elements on simplices.

This element's definition appears in https://doi.org/10.1007/BF02733251
(Regge, 1961), https://doi.org/10.1007/s00211-011-0394-z
(Christiansen, 2011), and
http://aurora.asc.tuwien.ac.at/~mneunteu/thesis/doctorthesis_neunteufel.pdf
(Neunteufel, 2021)
"""

import typing
from itertools import product

import sympy

from ..finite_element import CiarletElement
from ..functionals import (InnerProductIntegralMoment, IntegralAgainst, IntegralMoment,
                           ListOfFunctionals, PointInnerProduct)
from ..functions import FunctionInput
from ..moments import make_integral_moment_dofs
from ..polynomials import polynomial_set_vector
from ..references import Reference
from ..symbols import t, x
from .lagrange import Lagrange


class Regge(CiarletElement):
    """A Regge element on a simplex."""

    def __init__(self, reference: Reference, order: int, variant: str = "point"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        from symfem import create_reference
        poly: typing.List[FunctionInput] = []
        if reference.tdim == 2:
            poly = [((p[0], p[1]), (p[1], p[2]))
                    for p in polynomial_set_vector(reference.tdim, 3, order)]
        if reference.tdim == 3:
            poly = [((p[0], p[1], p[3]), (p[1], p[2], p[4]), (p[3], p[4], p[5]))
                    for p in polynomial_set_vector(reference.tdim, 6, order)]

        dofs: ListOfFunctionals = []
        if variant == "point":
            for edim in range(1, 4):
                et = reference.sub_entity_types[edim]
                for e_n in range(reference.sub_entity_count(edim)):
                    assert isinstance(et, str)
                    entity = reference.sub_entity(edim, e_n)
                    for i in product(range(1, order + 2), repeat=edim):
                        if sum(i) < order + 2:
                            for edge in entity.edges[::-1]:
                                tangent = tuple(b - a for a, b in zip(
                                    entity.vertices[edge[0]], entity.vertices[edge[1]]))
                                dofs.append(PointInnerProduct(
                                    reference, tuple(o + sum(sympy.Rational(a[j] * b, order + 2)
                                                             for a, b in zip(entity.axes, i[::-1]))
                                                     for j, o in enumerate(entity.origin)),
                                    tangent, tangent, entity=(edim, e_n),
                                    mapping="double_covariant"))

        elif variant == "integral":
            space = Lagrange(create_reference("interval"), order, "equispaced")
            basis = [f.subs(x, t) for f in space.get_basis_functions()]
            for e_n, vs in enumerate(reference.sub_entities(1)):
                edge_e = reference.sub_entity(1, e_n)
                tangent = tuple((b - a) / edge_e.jacobian()
                                for a, b in zip(edge_e.vertices[0], edge_e.vertices[1]))
                for f, dof in zip(basis, space.dofs):
                    dofs.append(InnerProductIntegralMoment(
                        reference, edge_e, f, tangent, tangent, dof, entity=(1, e_n),
                        mapping="double_covariant"))

            if reference.tdim == 2:
                if order > 0:
                    dofs += make_integral_moment_dofs(
                        reference,
                        cells=(IntegralMoment, Regge, order - 1, "double_covariant",
                               {"variant": "integral"}),
                    )

            elif reference.tdim == 3:
                if order > 0:
                    rspace = Regge(create_reference("triangle"), order - 1, "integral")
                    basis = [f.subs(x, t) for f in rspace.get_basis_functions()]
                    for f_n, vs in enumerate(reference.sub_entities(2)):
                        face = reference.sub_entity(2, f_n)
                        for f, dof in zip(basis, rspace.dofs):
                            dofs.append(IntegralMoment(
                                reference, face, f * face.jacobian(), dof,
                                entity=(2, f_n), mapping="double_covariant"))

                if order > 1:
                    dofs += make_integral_moment_dofs(
                        reference,
                        cells=(IntegralMoment, Regge, order - 2, "double_covariant",
                               {"variant": "integral"}),
                    )

        else:
            raise ValueError(f"Unknown variant: {variant}")

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim ** 2,
                         (reference.tdim, reference.tdim))
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["Regge"]
    references = ["triangle", "tetrahedron"]
    min_order = 0
    continuity = "inner H(curl)"


class ReggeTP(CiarletElement):
    """A Regge element on a tensor product cell."""

    def __init__(self, reference: Reference, order: int, variant: str = "integral"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        from symfem import create_reference

        poly: typing.List[FunctionInput] = []
        if reference.tdim == 2:
            for i in range(order + 1):
                for j in range(order + 2):
                    poly.append(((x[0] ** i * x[1] ** j, 0), (0, 0)))
                    poly.append(((0, 0), (0, x[0] ** j * x[1] ** i)))
            for i in range(order + 1):
                for j in range(order + 1):
                    poly.append(((0, x[0] ** i * x[1] ** j), (x[0] ** i * x[1] ** j, 0)))
        elif reference.tdim == 3:
            for i in range(order + 1):
                for j in range(order + 2):
                    for k in range(order + 2):
                        poly.append(
                            ((x[0] ** i * x[1] ** j * x[2] ** k, 0, 0), (0, 0, 0), (0, 0, 0)))
                        poly.append(
                            ((0, 0, 0), (0, x[1] ** i * x[0] ** j * x[2] ** k, 0), (0, 0, 0)))
                        poly.append(
                            ((0, 0, 0), (0, 0, 0), (0, 0, x[2] ** i * x[0] ** j * x[1] ** k)))
            for i in range(order + 1):
                for j in range(order + 1):
                    for k in range(order + 2):
                        poly.append(((0, x[0] ** i * x[1] ** j * x[2] ** k, 0),
                                     (x[0] ** i * x[1] ** j * x[2] ** k, 0, 0),
                                     (0, 0, 0)))
                        poly.append(((0, 0, x[0] ** i * x[2] ** j * x[1] ** k),
                                     (0, 0, 0),
                                     (x[0] ** i * x[2] ** j * x[1] ** k, 0, 0)))
                        poly.append(((0, 0, 0),
                                     (0, 0, x[1] ** i * x[2] ** j * x[0] ** k),
                                     (0, x[1] ** i * x[2] ** j * x[0] ** k, 0)))

        dofs: ListOfFunctionals = []
        if variant == "integral":
            # DOFs on edges
            space = Lagrange(create_reference("interval"), order, "equispaced")
            basis = [f.subs(x, t) for f in space.get_basis_functions()]
            for e_n, vs in enumerate(reference.sub_entities(1)):
                edge = reference.sub_entity(1, e_n)
                tangent = tuple((b - a) / edge.jacobian()
                                for a, b in zip(edge.vertices[0], edge.vertices[1]))
                for f, dof in zip(basis, space.dofs):
                    dofs.append(InnerProductIntegralMoment(
                        reference, edge, f, tangent, tangent, dof, entity=(1, e_n),
                        mapping="double_covariant"))

            # DOFs on faces
            for f_n, vs in enumerate(reference.sub_entities(2)):
                face = reference.sub_entity(2, f_n)
                for i in range(order + 1):
                    for j in range(order + 1):
                        dofs.append(IntegralAgainst(
                            reference, face,
                            ((0, x[0] ** i * x[1] ** j), (x[0] ** i * x[1] ** j, 0)),
                            entity=(2, f_n), mapping="double_covariant"))
                for i in range(1, order + 1):
                    for j in range(order + 1):
                        dofs.append(IntegralAgainst(
                            reference, face, ((x[1] ** i * x[0] ** j * (1 - x[1]), 0), (0, 0)),
                            entity=(2, f_n), mapping="double_covariant"))
                        dofs.append(IntegralAgainst(
                            reference, face, ((0, 0), (0, x[0] ** i * x[1] ** j * (1 - x[0]))),
                            entity=(2, f_n), mapping="double_covariant"))

            if reference.tdim == 3:
                # DOFs on cell
                for i in range(1, order + 1):
                    for j in range(order + 1):
                        for k in range(order + 1):
                            f = x[0] ** i * x[1] ** j * x[2] ** k * (1 - x[0])
                            assert isinstance(f, sympy.core.expr.Expr)
                            dofs.append(IntegralAgainst(
                                reference, reference, ((0, 0, 0), (0, 0, f), (0, f, 0)),
                                entity=(3, 0), mapping="double_covariant"))
                            f = x[1] ** i * x[0] ** j * x[2] ** k * (1 - x[1])
                            assert isinstance(f, sympy.core.expr.Expr)
                            dofs.append(IntegralAgainst(
                                reference, reference, ((0, 0, f), (0, 0, 0), (f, 0, 0)),
                                entity=(3, 0), mapping="double_covariant"))
                            f = x[2] ** i * x[0] ** j * x[1] ** k * (1 - x[2])
                            assert isinstance(f, sympy.core.expr.Expr)
                            dofs.append(IntegralAgainst(
                                reference, reference, ((0, f, 0), (f, 0, 0), (0, 0, 0)),
                                entity=(3, 0), mapping="double_covariant"))
                for i in range(order + 1):
                    for j in range(1, order + 1):
                        for k in range(1, order + 1):
                            f = x[0] ** i * x[1] ** j * x[2] ** k * (1 - x[1]) * (1 - x[2])
                            assert isinstance(f, sympy.core.expr.Expr)
                            dofs.append(IntegralAgainst(
                                reference, reference, ((f, 0, 0), (0, 0, 0), (0, 0, 0)),
                                entity=(3, 0), mapping="double_covariant"))
                            f = x[1] ** i * x[0] ** j * x[2] ** k * (1 - x[0]) * (1 - x[2])
                            assert isinstance(f, sympy.core.expr.Expr)
                            dofs.append(IntegralAgainst(
                                reference, reference, ((0, 0, 0), (0, f, 0), (0, 0, 0)),
                                entity=(3, 0), mapping="double_covariant"))
                            f = x[2] ** i * x[0] ** j * x[1] ** k * (1 - x[0]) * (1 - x[1])
                            assert isinstance(f, sympy.core.expr.Expr)
                            dofs.append(IntegralAgainst(
                                reference, reference, ((0, 0, 0), (0, 0, 0), (0, 0, f)),
                                entity=(3, 0), mapping="double_covariant"))
        else:
            raise ValueError(f"Unknown variant: {variant}")

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim ** 2,
                         (reference.tdim, reference.tdim))
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["Regge"]
    references = ["quadrilateral", "hexahedron"]
    min_order = 0
    continuity = "inner H(curl)"



"""Guzman-Neilan elements on simplices.

This element's definition appears in https://doi.org/10.1137/17M1153467
(Guzman and Neilan, 2018)
"""

import typing

import sympy

from ..finite_element import CiarletElement
from ..functionals import DotPointEvaluation, ListOfFunctionals, NormalIntegralMoment
from ..functions import FunctionInput, VectorFunction
from ..geometry import SetOfPoints, SetOfPointsInput
from ..moments import make_integral_moment_dofs
from ..piecewise_functions import PiecewiseFunction
from ..references import Reference
from .bernardi_raugel import BernardiRaugel
from .lagrange import Lagrange, VectorLagrange


class GuzmanNeilan(CiarletElement):
    """Guzman-Neilan Hdiv finite element."""

    def __init__(self, reference: Reference, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
        """
        if reference.name == "triangle":
            poly = self._make_polyset_triangle(reference, order)
        else:
            poly = self._make_polyset_tetrahedron(reference, order)

        dofs: ListOfFunctionals = []

        for n in range(reference.sub_entity_count(reference.tdim - 1)):
            facet = reference.sub_entity(reference.tdim - 1, n)
            for v in facet.vertices:
                dofs.append(DotPointEvaluation(
                    reference, v, tuple(i * facet.jacobian() for i in facet.normal()),
                    entity=(reference.tdim - 1, n),
                    mapping="contravariant"))

        if order == 2:
            assert reference.name == "tetrahedron"

            # Midpoints of edges
            for n in range(reference.sub_entity_count(1)):
                edge = reference.sub_entity(1, n)
                dofs.append(DotPointEvaluation(
                    reference, edge.midpoint(), tuple(i * edge.jacobian() for i in edge.tangent()),
                    entity=(1, n), mapping="contravariant"))

            # Midpoints of edges of faces
            for n in range(reference.sub_entity_count(2)):
                face = reference.sub_entity(2, n)
                for m in range(3):
                    edge = face.sub_entity(1, m)
                    dofs.append(DotPointEvaluation(
                        reference, edge.midpoint(),
                        tuple(i * face.jacobian() for i in face.normal()),
                        entity=(2, n), mapping="contravariant"))

            # Interior edges
            for v in reference.vertices:
                p = tuple((i + sympy.Rational(1, 4)) / 2 for i in v)
                for d in [(1, 0, 0), (0, 1, 0), (0, 0, 1)]:
                    dofs.append(DotPointEvaluation(
                        reference, p, d, entity=(3, 0), mapping="contravariant"))

        dofs += make_integral_moment_dofs(
            reference,
            facets=(NormalIntegralMoment, Lagrange, 0, "contravariant"),
        )

        mid = reference.midpoint()
        for i in range(reference.tdim):
            direction = tuple(1 if i == j else 0 for j in range(reference.tdim))
            dofs.append(DotPointEvaluation(reference, mid, direction, entity=(reference.tdim, 0),
                                           mapping="contravariant"))

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)

    def _make_polyset_triangle(
        self, reference: Reference, order: int
    ) -> typing.List[FunctionInput]:
        """Make the polyset for a triangle.

        Args:
            reference: The reference cell
            order: The polynomial order

        Returns:
            The polynomial set
        """
        assert order == 1

        from ._guzman_neilan_triangle import coeffs

        mid = reference.midpoint()

        sub_tris: typing.List[SetOfPoints] = [
            (reference.vertices[0], reference.vertices[1], mid),
            (reference.vertices[0], reference.vertices[2], mid),
            (reference.vertices[1], reference.vertices[2], mid)]

        basis: typing.List[FunctionInput] = []
        basis += make_piecewise_lagrange(sub_tris, "triangle", order)

        sub_basis = make_piecewise_lagrange(sub_tris, "triangle", reference.tdim, True)
        fs = BernardiRaugel(reference, 1).get_basis_functions()[-3:]
        for c, f in zip(coeffs, fs):
            assert isinstance(f, VectorFunction)
            pieces: typing.Dict[SetOfPointsInput, FunctionInput] = {}
            for tri in sub_tris:
                function: typing.List[sympy.core.expr.Expr] = []
                for j in range(reference.tdim):
                    component = f[j]
                    for k, b in zip(c, sub_basis):
                        component -= k * b.pieces[tri][j]
                    c_sym = component.as_sympy()
                    assert isinstance(c_sym, sympy.core.expr.Expr)
                    function.append(c_sym)
                pieces[tri] = VectorFunction(tuple(function))
            basis.append(PiecewiseFunction(pieces, 2))
        return basis

    def _make_polyset_tetrahedron(
        self, reference: Reference, order: int
    ) -> typing.List[FunctionInput]:
        """Make the polyset for a tetrahedron.

        Args:
            reference: The reference cell
            order: The polynomial order

        Returns:
            The polynomial set
        """
        assert order in [1, 2]
        from ._guzman_neilan_tetrahedron import coeffs

        mid = reference.midpoint()

        sub_tets: typing.List[SetOfPoints] = [
            (reference.vertices[0], reference.vertices[1], reference.vertices[2], mid),
            (reference.vertices[0], reference.vertices[1], reference.vertices[3], mid),
            (reference.vertices[0], reference.vertices[2], reference.vertices[3], mid),
            (reference.vertices[1], reference.vertices[2], reference.vertices[3], mid)]

        basis: typing.List[FunctionInput] = []
        basis += make_piecewise_lagrange(sub_tets, "tetrahedron", order)

        sub_basis = make_piecewise_lagrange(sub_tets, "tetrahedron", reference.tdim, True)
        fs = BernardiRaugel(reference, 1).get_basis_functions()[-4:]
        for c, f in zip(coeffs, fs):
            assert isinstance(f, VectorFunction)
            pieces: typing.Dict[SetOfPointsInput, FunctionInput] = {}
            for tet in sub_tets:
                function: typing.List[sympy.core.expr.Expr] = []
                for j in range(reference.tdim):
                    component = f[j]
                    for k, b in zip(c, sub_basis):
                        component -= k * b.pieces[tet][j]
                    c_sym = component.as_sympy()
                    assert isinstance(c_sym, sympy.core.expr.Expr)
                    function.append(c_sym)
                pieces[tet] = VectorFunction(tuple(function))
            basis.append(PiecewiseFunction(pieces, 3))
        return basis

    names = ["Guzman-Neilan"]
    references = ["triangle", "tetrahedron"]
    min_order = 1
    max_order = {"triangle": 1, "tetrahedron": 2}
    continuity = "H(div)"


def make_piecewise_lagrange(
    sub_cells: typing.List[SetOfPoints], cell_name, order: int, zero_on_boundary: bool = False,
    zero_at_centre: bool = False
) -> typing.List[PiecewiseFunction]:
    """Make the basis functions of a piecewise Lagrange space.

    Args:
        sub_cells: A list of vertices of sub cells
        cell_name: The cell type of the sub cells
        order: The polynomial order
        zero_in_boundary: Should the functions be zero on the boundary?
        zero_at_centre: Should the functions be zero at the centre?

    Returns:
        The basis functions
    """
    from symfem import create_reference
    lagrange_space = VectorLagrange(create_reference(cell_name), order)
    lagrange_bases: typing.List[typing.List[VectorFunction]] = []
    for c in sub_cells:
        row: typing.List[VectorFunction] = []
        c_basis = lagrange_space.map_to_cell(c)
        for cb in c_basis:
            assert isinstance(cb, VectorFunction)
            row.append(cb)
        lagrange_bases.append(row)

    basis_dofs: typing.List[typing.Tuple[int, ...]] = []
    zero: typing.Tuple[int, ...] = (0, )
    if cell_name == "triangle":
        cell_tdim = 2
        for dim, tri_entities in enumerate([
            [(0, 0, -1), (1, -1, 0), (-1, 1, 1), (2, 2, 2)],
            [(0, -1, 1), (1, 1, -1), (-1, 0, 0),
             (2, -1, -1), (-1, 2, -1), (-1, -1, 2)],
            [(0, -1, -1), (-1, 0, -1), (-1, -1, 0)]
        ]):
            nones = [
                -1 for i in lagrange_space.entity_dofs(dim, 0)]
            for tri_e in tri_entities:
                if dim == 0:
                    if zero_on_boundary and (0 in tri_e or 1 in tri_e):
                        continue
                    if zero_at_centre and (2 in tri_e):
                        continue
                elif dim == 1:
                    if zero_on_boundary and (2 in tri_e):
                        continue
                doflist = [
                    nones if i == -1 else lagrange_space.entity_dofs(dim, i) for i in tri_e
                ]
                for dofs in zip(*doflist):
                    basis_dofs.append(dofs)
        zero = (0, 0)

    elif cell_name == "tetrahedron":
        cell_tdim = 3
        for dim, tet_entities in enumerate([
            [(0, 0, 0, -1), (1, 1, -1, 0), (2, -1, 1, 1), (-1, 2, 2, 2), (3, 3, 3, 3)],
            [(2, -1, -1, 5), (5, 5, -1, -1), (4, -1, 5, -1), (-1, 2, -1, 4),
             (-1, 4, 4, -1), (-1, -1, 2, 2), (3, 3, 3, -1), (-1, 0, 0, 0), (0, -1, 1, 1),
             (1, 1, -1, 3)],
            [(0, -1, -1, 2), (1, -1, 2, -1), (2, 2, -1, -1), (3, -1, -1, -1), (-1, 0, -1, 1),
             (-1, 1, 1, -1), (-1, 3, -1, -1), (-1, -1, 0, 0), (-1, -1, 3, -1), (-1, -1, -1, 3)],
            [(0, -1, -1, -1), (-1, 0, -1, -1), (-1, -1, 0, -1), (-1, -1, -1, 0)]
        ]):
            nones = [
                -1 for i in lagrange_space.entity_dofs(dim, 0)]
            for tet_e in tet_entities:
                if dim == 0:
                    if zero_on_boundary and (0 in tet_e or 1 in tet_e or 2 in tet_e):
                        continue
                    if zero_at_centre and (3 in tet_e):
                        continue
                elif dim == 1:
                    if zero_on_boundary and (2 in tet_e or 4 in tet_e or 5 in tet_e):
                        continue
                elif dim == 2:
                    if zero_on_boundary and (3 in tet_e):
                        continue
                doflist = [
                    nones if i == -1 else lagrange_space.entity_dofs(dim, i) for i in tet_e
                ]
                for dofs in zip(*doflist):
                    basis_dofs.append(dofs)
        zero = (0, 0, 0)

    basis: typing.List[PiecewiseFunction] = []
    for i in basis_dofs:
        pieces: typing.Dict[SetOfPointsInput, FunctionInput] = {i: zero for i in sub_cells}
        for c, s, j in zip(sub_cells, lagrange_bases, i):
            if j != -1:
                pieces[c] = s[j]
        basis.append(PiecewiseFunction(pieces, cell_tdim))

    return basis



"""Test that basis functions agree with Basix."""

import typing

import pytest

from symfem import create_element

elements: typing.Dict[str, typing.List[typing.Tuple[
    str, str, typing.Iterable, typing.List[typing.Tuple[str, typing.Any]]
]]] = {
    "interval": [
        ("P", "Lagrange", range(1, 4), [("LagrangeVariant", "equispaced")]),
        ("serendipity", "Serendipity", range(1, 5), [("LagrangeVariant", "equispaced"),
                                                     ("DPCVariant", "simplex_equispaced")]),
        ("bubble", "Bubble", range(2, 5), []),
        ("dPc", "DPC", range(0, 5), [("LagrangeVariant", "equispaced"), ("bool", True)])],
    "triangle": [
        ("P", "Lagrange", range(1, 4), [("LagrangeVariant", "equispaced")]),
        ("bubble", "Bubble", range(3, 5), []),
        ("N1curl", "Nedelec 1st kind H(curl)", range(1, 4), [("LagrangeVariant", "equispaced")]),
        ("N2curl", "Nedelec 2nd kind H(curl)", range(1, 4), [("LagrangeVariant", "equispaced")]),
        ("N1div", "Raviart-Thomas", range(1, 4), [("LagrangeVariant", "equispaced")]),
        ("N2div", "Brezzi-Douglas-Marini", range(1, 4), [("LagrangeVariant", "equispaced")]),
        # ("Regge", "Regge", range(0, 4), []),
        # ("HHJ", "Hellan-Herrmann-Johnson", range(0, 4), []),
        ("Crouzeix-Raviart", "Crouzeix-Raviart", [1], [])],
    "tetrahedron": [
        ("P", "Lagrange", range(1, 4), [("LagrangeVariant", "equispaced")]),
        ("bubble", "Bubble", range(4, 6), []),
        ("N1curl", "Nedelec 1st kind H(curl)", range(1, 4), [("LagrangeVariant", "equispaced")]),
        ("N2curl", "Nedelec 2nd kind H(curl)", range(1, 3), [("LagrangeVariant", "equispaced")]),
        ("N1div", "Raviart-Thomas", range(1, 3), [("LagrangeVariant", "equispaced")]),
        ("N2div", "Brezzi-Douglas-Marini", range(1, 3), [("LagrangeVariant", "equispaced")]),
        # ("Regge", "Regge", range(0, 3), []),
        ("Crouzeix-Raviart", "Crouzeix-Raviart", [1], [])],
    "quadrilateral": [
        ("Q", "Lagrange", range(1, 4), [("LagrangeVariant", "equispaced")]),
        ("dPc", "DPC", range(0, 4), [("DPCVariant", "simplex_equispaced"), ("bool", True)]),
        ("serendipity", "Serendipity", range(1, 5),
         [("LagrangeVariant", "equispaced"), ("DPCVariant", "simplex_equispaced")]),
        ("Qdiv", "Raviart-Thomas", range(1, 4), [("LagrangeVariant", "equispaced")]),
        ("Qcurl", "Nedelec 1st kind H(curl)", range(1, 4), [("LagrangeVariant", "equispaced")]),
        ("Sdiv", "Brezzi-Douglas-Marini", range(1, 4),
         [("LagrangeVariant", "equispaced"), ("DPCVariant", "simplex_equispaced")]),
        ("Scurl", "Nedelec 2nd kind H(curl)", range(1, 4),
         [("LagrangeVariant", "equispaced"), ("DPCVariant", "simplex_equispaced")])],
    "hexahedron": [
        ("Q", "Lagrange", range(1, 3), [("LagrangeVariant", "equispaced")]),
        ("dPc", "DPC", range(0, 3),
         [("DPCVariant", "simplex_equispaced"), ("bool", True)]),
        ("serendipity", "Serendipity", range(1, 5),
         [("LagrangeVariant", "equispaced"), ("DPCVariant", "simplex_equispaced")]),
        ("Qdiv", "Raviart-Thomas", range(1, 3), [("LagrangeVariant", "equispaced")]),
        ("Qcurl", "Nedelec 1st kind H(curl)", range(1, 3), [("LagrangeVariant", "equispaced")]),
        ("Sdiv", "Brezzi-Douglas-Marini", range(1, 3),
         [("LagrangeVariant", "equispaced"), ("DPCVariant", "simplex_equispaced")]),
        ("Scurl", "Nedelec 2nd kind H(curl)", range(1, 3),
         [("LagrangeVariant", "equispaced"), ("DPCVariant", "simplex_equispaced")])],
    "prism": [
        ("Lagrange", "Lagrange", range(1, 4), [("LagrangeVariant", "equispaced")])]
}


def to_float(a):
    try:
        return float(a)
    except:  # noqa: E722
        return [to_float(i) for i in a]


def to_nparray(a):
    import numpy as np
    try:
        return float(a)
    except:  # noqa: E722
        return np.array([to_float(i) for i in a])


def make_lattice(cell, N=3):
    import numpy as np
    if cell == "interval":
        return np.array([[i / N] for i in range(N + 1)])
    if cell == "triangle":
        return np.array([[i / N, j / N] for i in range(N + 1) for j in range(N + 1 - i)])
    if cell == "tetrahedron":
        return np.array([[i / N, j / N, k / N]
                         for i in range(N + 1) for j in range(N + 1 - i)
                         for k in range(N + 1 - i - j)])
    if cell == "quadrilateral":
        return np.array([[i / N, j / N] for i in range(N + 1) for j in range(N + 1)])
    if cell == "hexahedron":
        return np.array([[i / N, j / N, k / N]
                         for i in range(N + 1) for j in range(N + 1) for k in range(N + 1)])
    if cell == "prism":
        return np.array([[i / N, j / N, k / N]
                         for i in range(N + 1) for j in range(N + 1 - i) for k in range(N + 1)])
    if cell == "pyramid":
        return np.array([[i / N, j / N, k / N]
                         for i in range(N + 1) for j in range(N + 1)
                         for k in range(N + 1 - max(i, j))])


@pytest.mark.parametrize(("cell", "symfem_type", "basix_type", "order", "args"),
                         [(cell, a, b, order, args) for cell, ls in elements.items()
                          for a, b, orders, args in ls for order in orders])
def test_against_basix(has_basix, elements_to_test, cells_to_test, cell, symfem_type,
                       basix_type, order, args, speed):
    if elements_to_test != "ALL" and symfem_type not in elements_to_test:
        pytest.skip()
    if cells_to_test != "ALL" and cell not in cells_to_test:
        pytest.skip()
    if speed == "fast" and order > 2:
        pytest.skip()

    # TODO: Implement faster non-symbolic mode and remove this
    if order > 1:
        pytest.skip()

    if symfem_type in ["Sdiv", "Scurl"]:
        pytest.xfail("Basix elements cannot yet be provided equispaced variant")

    if has_basix:
        import basix
    else:
        try:
            import basix
        except ImportError:
            pytest.skip("Basix must be installed to run this test.")

    import numpy as np

    points = make_lattice(cell, 2)
    parsed_args = []
    for a in args:
        if a[0] == "LagrangeVariant":
            parsed_args.append(basix.variants.string_to_lagrange_variant(a[1]))
        elif a[0] == "DPCVariant":
            parsed_args.append(basix.variants.string_to_dpc_variant(a[1]))
        elif a[0] == "bool":
            parsed_args.append(a[1])
        else:
            raise ValueError(f"Unknown arg type: {a[0]}")
    space = basix.create_element(
        basix.finite_element.string_to_family(basix_type, cell),
        basix.cell.string_to_type(cell), order, *parsed_args)
    result = space.tabulate(0, points)[0]

    element = create_element(cell, symfem_type, order)
    sym_result = to_nparray(element.tabulate_basis(points, "xyz,xyz"))

    if len(result.shape) != len(sym_result.shape):
        sym_result = sym_result.reshape(result.shape)

    assert np.allclose(result, sym_result)



"""Huang-Zhang element on a quadrilateral.

This element's definition appears in https://doi.org/10.1007/s11464-011-0094-0
(Huang, Zhang, 2011) and https://doi.org/10.1137/080728949 (Zhang, 2009)
"""

import typing

from ..finite_element import CiarletElement
from ..functionals import (IntegralAgainst, ListOfFunctionals, NormalIntegralMoment,
                           TangentIntegralMoment)
from ..functions import FunctionInput, VectorFunction
from ..moments import make_integral_moment_dofs
from ..references import Reference
from ..symbols import x
from .lagrange import Lagrange


class HuangZhang(CiarletElement):
    """Huang-Zhang finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        assert reference.name == "quadrilateral"
        self.variant = variant

        dofs: ListOfFunctionals = []
        poly: typing.List[FunctionInput] = []
        poly += [
            VectorFunction([x[0] ** i * x[1] ** j, 0])
            for i in range(order + 1)
            for j in range(order)
        ]
        poly += [
            VectorFunction([0, x[0] ** i * x[1] ** j])
            for i in range(order)
            for j in range(order + 1)
        ]

        dofs += make_integral_moment_dofs(
            reference,
            facets=(NormalIntegralMoment, Lagrange, order - 1,
                    {"variant": variant}),
        )
        dofs += make_integral_moment_dofs(
            reference,
            facets=(TangentIntegralMoment, Lagrange, order - 2,
                    {"variant": variant}),
        )

        for i in range(order - 1):
            for j in range(order - 2):
                dofs.append(IntegralAgainst(
                    reference, reference, (x[0] ** i * x[1] ** j, 0), (2, 0)))
                dofs.append(IntegralAgainst(
                    reference, reference, (0, x[0] ** j * x[1] ** i), (2, 0)))

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["Huang-Zhang", "HZ"]
    references = ["quadrilateral"]
    min_order = 2
    continuity = "H(div)"



"""Dual elements.

These elements' definitions appear in https://doi.org/10.1016/j.crma.2004.12.022
(Buffa, Christiansen, 2005)
"""

import typing

import sympy

from ..finite_element import FiniteElement
from ..functions import AnyFunction, FunctionInput, VectorFunction
from ..geometry import PointType, SetOfPoints, SetOfPointsInput
from ..piecewise_functions import PiecewiseFunction
from ..plotting import Picture, colors
from ..references import DualPolygon


class DualCiarletElement(FiniteElement):
    """Abstract barycentric finite element."""

    def __init__(
        self, dual_coefficients: typing.List[typing.List[typing.List[
            typing.Union[int, sympy.core.expr.Expr]]]],
        fine_space: str, reference: DualPolygon, order: int,
        dof_entities: typing.List[typing.Tuple[int, int]],
        domain_dim: int, range_dim: int,
        range_shape: typing.Optional[typing.Tuple[int, ...]] = None,
        dof_directions: typing.Optional[SetOfPoints] = None
    ):
        """Create a dual element.

        Args:
            dual_coefficients: the coefficients that define this element in terms of the basis
                               functions of the fine space
            fine_space: the family of the fine space
            reference: The reference element
            order: The polynomia order of the fine space
            dof_entities: The cell entity that each basis function is associated with
            domain_dim: the topological dimension of the domain
            range_dim: the dimension of the range
            range_shape: the shape of the range
            dof_directions: The direction that each basis function is associated with
        """
        self.dual_coefficients = dual_coefficients
        self.fine_space = fine_space
        super().__init__(reference, order, len(dual_coefficients), domain_dim, range_dim,
                         range_shape=range_shape)
        self._basis_functions: typing.Union[typing.List[AnyFunction], None] = None
        self.dof_entities = dof_entities
        self.dof_directions = dof_directions

    def get_polynomial_basis(
        self, reshape: bool = True
    ) -> typing.List[AnyFunction]:
        """Get the symbolic polynomial basis for the element.

        Returns:
            The polynomial basis
        """
        raise ValueError("Polynomial basis not supported for barycentric dual elements.")

    def get_dual_matrix(self) -> sympy.matrices.dense.MutableDenseMatrix:
        """Get the dual matrix.

        Returns:
            The dual matrix
        """
        raise ValueError("Dual matrix not supported for barycentric dual elements.")

    def get_basis_functions(
        self, use_tensor_factorisation: bool = False
    ) -> typing.List[AnyFunction]:
        """Get the basis functions of the element.

        Args:
            use_tensor_factorisation: Should a tensor factorisation be used?

        Returns:
            The basis functions
        """
        assert not use_tensor_factorisation

        if self._basis_functions is None:
            from symfem import create_element

            bfs: typing.List[AnyFunction] = []
            sub_e = create_element("triangle", self.fine_space, self.order)
            for coeff_list in self.dual_coefficients:
                v0 = self.reference.origin
                pieces: typing.Dict[SetOfPointsInput, FunctionInput] = {}
                for coeffs, v1, v2 in zip(
                    coeff_list, self.reference.vertices,
                    self.reference.vertices[1:] + self.reference.vertices[:1]
                ):
                    sub_basis = sub_e.map_to_cell((v0, v1, v2))

                    if self.range_dim == 1:
                        sub_fun = sympy.Integer(0)
                        for a, b in zip(coeffs, sub_basis):
                            sub_fun += a * b
                    else:
                        sf_list = []
                        for i in range(self.range_dim):
                            sf_item = sympy.Integer(0)
                            for a, b in zip(coeffs, sub_basis):
                                assert isinstance(b, VectorFunction)
                                sf_item += a * b[i]
                            sf_list.append(sf_item)
                        sub_fun = tuple(sf_list)
                    pieces[(v0, v1, v2)] = sub_fun
                bfs.append(PiecewiseFunction(pieces, 2))
            assert len(bfs) == len(self.dof_entities)
            self._basis_functions = bfs

        assert self._basis_functions is not None
        return self._basis_functions

    def entity_dofs(self, entity_dim: int, entity_number: int) -> typing.List[int]:
        """Get the numbers of the DOFs associated with the given entity.

        Args:
            entity_dim: The dimension of the entity
            entity_number: The number of the entity

        Returns:
            The numbers of the DOFs associated with the entity
        """
        out = []
        for i, e in enumerate(self.dof_entities):
            if e == (entity_dim, entity_number):
                out.append(i)
        return out

    def map_to_cell(
        self, vertices_in: SetOfPointsInput, basis:
        typing.Optional[typing.List[AnyFunction]] = None,
        forward_map: typing.Optional[PointType] = None,
        inverse_map: typing.Optional[PointType] = None
    ) -> typing.List[AnyFunction]:
        """Map the basis onto a cell using the appropriate mapping for the element.

        Args:
            vertices_in: The vertices of the cell
            basis: The basis functions
            forward_map: The map from the reference to the cell
            inverse_map: The map to the reference from the cell

        Returns:
            The basis functions mapped to the cell
        """
        raise NotImplementedError()

    def plot_dof_diagram(
        self, filename: typing.Union[str, typing.List[str]],
        plot_options: typing.Dict[str, typing.Any] = {}, **kwargs: typing.Any
    ):
        """Plot a diagram showing the DOFs of the element.

        Args:
            filename: The file name
            plot_options: Options for the plot
            kwargs: Keyword arguments
        """
        img = Picture(**kwargs)

        for entities in self.reference.z_ordered_entities():
            for dim, e_n in entities:
                if dim == 1:
                    pts = tuple(self.reference.vertices[i] for i in self.reference.edges[e_n])
                    img.add_line(pts[0], pts[1], colors.BLACK)

            for dim, e_n in entities:
                for d in self.entity_dofs(dim, e_n):
                    if dim == 0:
                        point = self.reference.vertices[e_n]
                    elif dim == 1:
                        point = tuple((a + b) / 2 for a, b in zip(
                            self.reference.vertices[self.reference.edges[e_n][0]],
                            self.reference.vertices[self.reference.edges[e_n][1]],
                        ))
                    elif dim == 2:
                        point = self.reference.midpoint()
                    else:
                        raise ValueError("Unsupported tdim")

                    if self.dof_directions is not None:
                        direction = self.dof_directions[d]
                        img.add_dof_arrow(point, direction, d, colors.entity(dim), False)
                    else:
                        img.add_dof_marker(point, d, colors.entity(dim))

        img.save(filename, plot_options=plot_options)


class Dual(DualCiarletElement):
    """Barycentric dual finite element."""

    def __init__(self, reference: DualPolygon, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
        """
        dual_coefficients: typing.List[typing.List[typing.List[
            typing.Union[int, sympy.core.expr.Expr]]]] = []
        if order == 0:
            dual_coefficients = [
                [[1] for i in range(2 * reference.number_of_triangles)]
            ]
            fine_space = "Lagrange"
            dof_entities = [(2, 0)]
        else:
            dual_coefficients = [
                [[sympy.Rational(1, reference.number_of_triangles), 0, 0]
                 for i in range(2 * reference.number_of_triangles)]
                for j in range(reference.number_of_triangles)
            ]

            for j in range(reference.number_of_triangles):
                dual_coefficients[j][2 * j][2] = 1
                dual_coefficients[j][2 * j + 1][1] = 1
                dual_coefficients[j][2 * j - 1][2] = sympy.Rational(1, 2)
                dual_coefficients[j][2 * j][1] = sympy.Rational(1, 2)
                dual_coefficients[j][2 * j + 1][2] = sympy.Rational(1, 2)
                if j + 1 == reference.number_of_triangles:
                    dual_coefficients[j][0][1] = sympy.Rational(1, 2)
                else:
                    dual_coefficients[j][2 * j + 2][1] = sympy.Rational(1, 2)

            dof_entities = [(1, i) for i in range(1, len(reference.vertices), 2)]

            fine_space = "Lagrange"

        super().__init__(
            dual_coefficients, fine_space, reference, order, dof_entities, reference.tdim, 1
        )

    names = ["dual polynomial", "dual P", "dual"]
    references = ["dual polygon"]
    min_order = 0
    max_order = 1
    continuity = "C0"


class BuffaChristiansen(DualCiarletElement):
    """Buffa-Christiansen barycentric dual finite element."""

    def __init__(self, reference: DualPolygon, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
        """
        assert order == 1
        dual_coefficients: typing.List[typing.List[typing.List[
            typing.Union[int, sympy.core.expr.Expr]]]] = [
            [[0, 0, 0]
             for i in range(2 * reference.number_of_triangles)]
            for j in range(reference.number_of_triangles)
        ]

        for j in range(reference.number_of_triangles):
            dual_coefficients[j][2 * j][0] = sympy.Rational(-1, 2)
            dual_coefficients[j][2 * j - 1][0] = sympy.Rational(-1, 2)
            N = 2 * reference.number_of_triangles
            for i in range(N - 1):
                dual_coefficients[j][(2 * j + i) % N][2] = sympy.Rational(i + 1 - N // 2, N)
                dual_coefficients[j][(2 * j + i + 1) % N][1] = sympy.Rational(i + 1 - N // 2, N)

        dof_entities = [(0, i) for i in range(0, len(reference.vertices), 2)]
        dof_directions: typing.List[PointType] = []
        for i in range(0, len(reference.vertices), 2):
            dof_directions.append(tuple(
                b - a for a, b in zip(reference.origin, reference.vertices[i])))

        super().__init__(
            dual_coefficients, "RT", reference, order, dof_entities, reference.tdim, 2,
            dof_directions=tuple(dof_directions))

    names = ["Buffa-Christiansen", "BC"]
    references = ["dual polygon"]
    min_order = 1
    max_order = 1
    continuity = "H(div)"


class RotatedBuffaChristiansen(DualCiarletElement):
    """RotatedBuffa-Christiansen barycentric dual finite element."""

    def __init__(self, reference: DualPolygon, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
        """
        assert order == 1
        dual_coefficients: typing.List[typing.List[typing.List[
            typing.Union[int, sympy.core.expr.Expr]]]] = [
            [[0, 0, 0]
             for i in range(2 * reference.number_of_triangles)]
            for j in range(reference.number_of_triangles)
        ]

        for j in range(reference.number_of_triangles):
            dual_coefficients[j][2 * j][0] = sympy.Rational(-1, 2)
            dual_coefficients[j][2 * j - 1][0] = sympy.Rational(-1, 2)
            N = 2 * reference.number_of_triangles
            for i in range(N - 1):
                dual_coefficients[j][(2 * j + i) % N][2] = sympy.Rational(N // 2 - 1 - i, N)
                dual_coefficients[j][(2 * j + i + 1) % N][1] = sympy.Rational(N // 2 - 1 - i, N)

        dof_entities = [(0, i) for i in range(0, len(reference.vertices), 2)]
        dof_directions: typing.List[PointType] = []
        for i in range(0, len(reference.vertices), 2):
            dof_directions.append(tuple(
                b - a for a, b in zip(reference.origin, reference.vertices[i])))

        super().__init__(
            dual_coefficients, "N1curl", reference, order, dof_entities,
            reference.tdim, 2, dof_directions=tuple(dof_directions))

    names = ["rotated Buffa-Christiansen", "RBC"]
    references = ["dual polygon"]
    min_order = 1
    max_order = 1
    continuity = "H(div)"



"""Morley elements on simplices.

This element's definition appears in https://doi.org/10.1017/S0001925900004546
(Morley, 1968)
"""

import typing

from ..finite_element import CiarletElement
from ..functionals import ListOfFunctionals, PointEvaluation, PointNormalDerivativeEvaluation
from ..functions import FunctionInput
from ..polynomials import polynomial_set_1d
from ..references import Reference


class Morley(CiarletElement):
    """Morley finite element."""

    def __init__(self, reference: Reference, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
        """
        if reference.vertices != reference.reference_vertices:
            raise NotImplementedError()
        assert order == 2
        assert reference.name == "triangle"
        dofs: ListOfFunctionals = []
        for v_n, vs in enumerate(reference.vertices):
            dofs.append(PointEvaluation(reference, vs, entity=(0, v_n)))
        for e_n in range(reference.sub_entity_count(1)):
            sub_ref = reference.sub_entity(1, e_n)
            midpoint = sub_ref.midpoint()
            dofs.append(
                PointNormalDerivativeEvaluation(reference, midpoint, sub_ref, entity=(1, e_n)))

        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_1d(reference.tdim, order)

        super().__init__(reference, order, poly, dofs, reference.tdim, 1)

    names = ["Morley"]
    references = ["triangle"]
    min_order = 2
    max_order = 2
    continuity = "L2"



"""Test Arnold-Winther element."""

import pytest

from symfem import create_element


@pytest.mark.parametrize("order", range(3, 7))
def test_create(order):
    create_element("triangle", "Arnold-Winther", order)



"""Bubble elements on simplices.

This element's definition appears in https://doi.org/10.1007/978-3-642-23099-8_3
(Kirby, Logg, Rognes, Terrel, 2012)
"""

import typing
from itertools import product

import sympy

from ..finite_element import CiarletElement
from ..functionals import DotPointEvaluation, ListOfFunctionals, PointEvaluation
from ..functions import FunctionInput
from ..polynomials import polynomial_set_1d, quolynomial_set_1d
from ..references import Reference
from .lagrange import Lagrange


class Bubble(CiarletElement):
    """Bubble finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        from .. import create_element
        p1 = create_element(reference.name, "Lagrange", 1)
        bubble = 1
        for f in p1.get_basis_functions():
            bubble *= f

        poly: typing.List[FunctionInput] = []

        # TODO: variants
        if reference.name == "interval":
            poly += [bubble * p for p in polynomial_set_1d(reference.tdim, order - 2)]
        elif reference.name == "triangle":
            poly += [bubble * p for p in polynomial_set_1d(reference.tdim, order - 3)]
        elif reference.name == "tetrahedron":
            poly += [bubble * p for p in polynomial_set_1d(reference.tdim, order - 4)]
        elif reference.name == "quadrilateral":
            poly += [bubble * p for p in quolynomial_set_1d(reference.tdim, order - 2)]
        elif reference.name == "hexahedron":
            poly += [bubble * p for p in quolynomial_set_1d(reference.tdim, order - 2)]
        else:
            raise ValueError(f"Unsupported reference: {reference.name}")

        dofs: ListOfFunctionals = []
        if reference.name in ["interval", "triangle", "tetrahedron"]:
            def func(i): return sum(i)
        else:
            def func(i): return max(i)
        for i in product(range(1, order), repeat=reference.tdim):
            if func(i) < order:
                point = tuple(sympy.Rational(j, order) for j in i)
                dofs.append(PointEvaluation(reference, point, entity=(reference.tdim, 0)))

        self.variant = variant

        super().__init__(
            reference, order, poly, dofs, reference.tdim, 1
        )

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["bubble"]
    references = ["interval", "triangle", "tetrahedron", "quadrilateral", "hexahedron"]
    min_order = {"interval": 2, "triangle": 3, "tetrahedron": 4,
                 "quadrilateral": 2, "hexahedron": 2}
    continuity = "C0"


class BubbleEnrichedLagrange(CiarletElement):
    """Bubble enriched Lagrange element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        lagrange = Lagrange(reference, order, variant)
        bubble = Bubble(reference, order + 2, variant)

        poly: typing.List[FunctionInput] = []
        for e in [lagrange, bubble]:
            for p in e._basis:
                poly.append(p)

        self.variant = variant

        super().__init__(
            reference, order, poly,
            lagrange.dofs + bubble.dofs, reference.tdim, 1
        )

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["bubble enriched Lagrange"]
    references = ["triangle"]
    min_order = 1
    continuity = "C0"


class BubbleEnrichedVectorLagrange(CiarletElement):
    """Bubble enriched Lagrange element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        lagrange = Lagrange(reference, order, variant)
        bubble = Bubble(reference, order + 2, variant)

        poly: typing.List[FunctionInput] = []
        for e in [lagrange, bubble]:
            for p in e._basis:
                poly.append((p, 0))
                poly.append((0, p))

        dofs: ListOfFunctionals = [DotPointEvaluation(reference, d.dof_point(), v, entity=d.entity)
                                   for d in lagrange.dofs + bubble.dofs
                                   for v in [(1, 0), (0, 1)]]

        self.variant = variant

        super().__init__(reference, order, poly, dofs, reference.tdim, 2)

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["bubble enriched vector Lagrange"]
    references = ["triangle"]
    min_order = 1
    continuity = "C0"



"""Functions to create integral moments."""

import typing

from .functionals import BaseFunctional
from .references import Reference

MomentType = typing.Tuple[typing.Type, typing.Type, int, typing.Union[str, None],
                          typing.Dict[str, typing.Any]]
SingleMomentTypeInput = typing.Union[
    MomentType,
    typing.Tuple[typing.Type, typing.Type, int, str],
    typing.Tuple[typing.Type, typing.Type, int, typing.Dict[str, typing.Any]],
    typing.Tuple[typing.Type, typing.Type, int],
]
MomentTypeInput = typing.Union[
    SingleMomentTypeInput,
    typing.Dict[str, SingleMomentTypeInput]
]


def _extract_moment_data(moment_data: MomentTypeInput, sub_type: str) -> MomentType:
    """Get the information for a moment.

    Args:
        moment_data: The moment data
        sub_type: The subentity type

    Returns:
        The moment type, finite elment, order, mapping, and keyword arguments for the moment
    """
    if isinstance(moment_data, dict):
        return _extract_moment_data(moment_data[sub_type], sub_type)

    mapping: typing.Union[str, None] = None
    if isinstance(moment_data[-1], dict):
        kwargs = moment_data[-1]
        if isinstance(moment_data[-2], str):
            mapping = moment_data[-2]
    else:
        kwargs = {}
        if isinstance(moment_data[-1], str):
            mapping = moment_data[-1]

    assert isinstance(moment_data[0], type)
    assert isinstance(moment_data[1], type)
    assert isinstance(moment_data[2], int)

    return moment_data[0], moment_data[1], moment_data[2], mapping, kwargs


def make_integral_moment_dofs(
    reference: Reference,
    vertices: typing.Optional[MomentTypeInput] = None,
    edges: typing.Optional[MomentTypeInput] = None,
    faces: typing.Optional[MomentTypeInput] = None,
    volumes: typing.Optional[MomentTypeInput] = None,
    cells: typing.Optional[MomentTypeInput] = None,
    facets: typing.Optional[MomentTypeInput] = None,
    ridges: typing.Optional[MomentTypeInput] = None,
    peaks: typing.Optional[MomentTypeInput] = None
) -> typing.List[BaseFunctional]:
    """Generate DOFs due to integral moments on sub entities.

    Args:
        reference: The reference cell.
        vertices: DOFs on dimension 0 entities.
        edges: DOFs on dimension 1 entities.
        faces: DOFs on dimension 2 entities.
        volumes: DOFs on dimension 3 entities.
        cells: DOFs on codimension 0 entities.
        facets: DOFs on codimension 1 entities.
        ridges: DOFs on codimension 2 entities.
        peaks: DOFs on codimension 3 entities.

    Returns:
        A list of DOFs for the element
    """
    dofs = []

    # DOFs per dimension
    for dim, moment_data in [
        (0, vertices), (1, edges), (2, faces), (3, volumes),
        (reference.tdim - 3, peaks), (reference.tdim - 2, ridges), (reference.tdim - 1, facets),
        (reference.tdim, cells),
    ]:
        if moment_data is not None:
            sub_type = reference.sub_entity_types[dim]
            if sub_type is not None:
                assert dim > 0
                for i, vs in enumerate(reference.sub_entities(dim)):
                    sub_ref = reference.sub_entity(dim, i, True)
                    IntegralMoment, SubElement, order, mapping, kwargs = _extract_moment_data(
                        moment_data, sub_ref.name)
                    if order >= SubElement.min_order:
                        sub_element = SubElement(sub_ref.default_reference(), order, **kwargs)
                        for dn, d in enumerate(sub_element.dofs):
                            f = sub_element.get_basis_function(dn)
                            kwargs = {}
                            if mapping is not None:
                                kwargs["mapping"] = mapping
                            dofs.append(IntegralMoment(
                                reference, sub_ref, f, d, (dim, i), **kwargs))
    return dofs



"""Crouzeix-Raviart elements on simplices.

This element's definition appears in https://doi.org/10.1051/m2an/197307R300331
(Crouzeix, Raviart, 1973)
"""

import typing
from itertools import product

from ..finite_element import CiarletElement
from ..functionals import ListOfFunctionals, PointEvaluation
from ..functions import FunctionInput
from ..polynomials import polynomial_set_1d
from ..quadrature import get_quadrature
from ..references import Reference


class CrouzeixRaviart(CiarletElement):
    """Crouzeix-Raviart finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        assert reference.name in ["triangle", "tetrahedron"]

        if order > 1:
            assert reference.name == "triangle"

        points, _ = get_quadrature(variant, order + reference.tdim)

        dofs: ListOfFunctionals = []

        for e_n in range(reference.sub_entity_count(reference.tdim - 1)):
            entity = reference.sub_entity(reference.tdim - 1, e_n)
            for i in product(range(1, order + 1), repeat=reference.tdim - 1):
                if sum(i) < order + reference.tdim - 1:
                    dofs.append(
                        PointEvaluation(
                            reference,
                            tuple(o + sum(a[j] * points[b]
                                          for a, b in zip(entity.axes, i))
                                  for j, o in enumerate(entity.origin)),
                            entity=(reference.tdim - 1, e_n)))

        points, _ = get_quadrature(variant, order + reference.tdim - 1)
        for i in product(range(1, order), repeat=reference.tdim):
            if sum(i) < order:
                dofs.append(
                    PointEvaluation(
                        reference,
                        tuple(o + sum(a[j] * points[b]
                                      for a, b in zip(reference.axes, i))
                              for j, o in enumerate(reference.origin)),
                        entity=(reference.tdim, 0)))

        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_1d(reference.tdim, order)

        self.variant = variant
        super().__init__(reference, order, poly, dofs, reference.tdim, 1)

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["Crouzeix-Raviart", "CR", "Crouzeix-Falk", "CF"]
    references = ["triangle", "tetrahedron"]
    min_order = 1
    max_order = {"tetrahedron": 1}
    continuity = "L2"



"""Define fixtures and other test helpers."""

import pytest


def pytest_addoption(parser):
    """Add parser options."""
    parser.addoption("--elements-to-test", action="store", default="ALL")
    parser.addoption("--cells-to-test", action="store", default="ALL")
    parser.addoption("--has-basix", action="store", default="0")
    parser.addoption("--speed", action="store", default="slow")


@pytest.fixture
def elements_to_test(request):
    """Get list of elements to include in tests."""
    data = request.config.getoption("--elements-to-test")
    if data == "ALL":
        return "ALL"
    return data.split(",")


@pytest.fixture
def speed(request):
    """Get test speed."""
    return request.config.getoption("--speed")


@pytest.fixture
def cells_to_test(request):
    """Get list of cells to include in tests."""
    data = request.config.getoption("--cells-to-test")
    if data == "ALL":
        return "ALL"
    return data.split(",")


@pytest.fixture
def has_basix(request):
    """Get has-basix flag."""
    data = request.config.getoption("--has-basix")
    return data == "1"



"""Symfem."""

import os
import sys

import setuptools

if sys.version_info < (3, 6):
    print("Python 3.6 or higher required, please upgrade.")
    sys.exit(1)

with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), "README.md")) as f:
    long_description = f.read().replace(
        "(logo/logo.png)",
        "(https://raw.githubusercontent.com/mscroggs/symfem/main/logo/logo.png)"
    ).replace(
        "(img/",
        "(https://raw.githubusercontent.com/mscroggs/symfem/main/img/")
with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), "README.md"), "w") as f:
    f.write(long_description)

data_files = ["LICENSE", "README.md",
              ("test", ["test/__init__.py", "test/utils.py", "test/conftest.py"])]

if __name__ == "__main__":
    setuptools.setup(
        name="symfem",
        description="a symbolic finite element definition library",
        long_description=long_description,
        long_description_content_type="text/markdown",
        version="2023.1.1",
        author="Matthew Scroggs",
        license="MIT",
        author_email="symfem@mscroggs.co.uk",
        maintainer_email="symfem@mscroggs.co.uk",
        url="https://github.com/mscroggs/symfem",
        packages=["symfem", "symfem.elements"],
        package_data={"symfem": ["py.typed"]},
        include_package_data=True,
        data_files=data_files,
        install_requires=["sympy>=1.10"],
        extras_require={
            "style": ["flake8", "pydocstyle", "mypy", "isort"],
            "docs": ["sphinx==5.0.2", "sphinx-autoapi"],
            "optional": ["CairoSVG"],
            "test": ["pytest", "symfem[optional]", "numpy"],
        }
    )



"""Argyris elements on simplices.

This element's definition appears in https://doi.org/10.1017/S000192400008489X
(Arygris, Fried, Scharpf, 1968)
"""

import typing

from ..finite_element import CiarletElement
from ..functionals import (ListOfFunctionals, PointComponentSecondDerivativeEvaluation,
                           PointDirectionalDerivativeEvaluation, PointEvaluation,
                           PointNormalDerivativeEvaluation)
from ..functions import FunctionInput
from ..polynomials import polynomial_set_1d
from ..references import Reference


class Argyris(CiarletElement):
    """Argyris finite element."""

    def __init__(self, reference: Reference, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
        """
        assert order == 5
        assert reference.name == "triangle"
        dofs: ListOfFunctionals = []
        for v_n, vs in enumerate(reference.vertices):
            dofs.append(PointEvaluation(reference, vs, entity=(0, v_n)))
            for i in range(reference.tdim):
                direction = tuple(1 if i == j else 0 for j in range(reference.tdim))
                dofs.append(PointDirectionalDerivativeEvaluation(
                    reference, vs, direction, entity=(0, v_n)))
            for i in range(reference.tdim):
                for j in range(i + 1):
                    dofs.append(PointComponentSecondDerivativeEvaluation(
                        reference, vs, (i, j), entity=(0, v_n)))
        for e_n in range(reference.sub_entity_count(1)):
            assert isinstance(reference.sub_entity_types[1], str)
            sub_ref = reference.sub_entity(1, e_n)
            dofs.append(PointNormalDerivativeEvaluation(
                reference, sub_ref.midpoint(), sub_ref, entity=(1, e_n)))
        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_1d(reference.tdim, order)

        super().__init__(reference, order, poly, dofs, reference.tdim, 1)

    names = ["Argyris"]
    references = ["triangle"]
    min_order = 5
    max_order = 5
    continuity = "L2"



"""Fortin-Soulie elements on a triangle.

This element's definition appears in https://doi.org/10.1002/nme.1620190405
(Fortin, Soulie, 1973)
"""

import typing

import sympy

from ..finite_element import CiarletElement
from ..functionals import ListOfFunctionals, PointEvaluation
from ..functions import FunctionInput
from ..polynomials import polynomial_set_1d
from ..references import Reference


class FortinSoulie(CiarletElement):
    """Fortin-Soulie finite element."""

    def __init__(self, reference: Reference, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
        """
        assert reference.name == "triangle"

        assert order == 2

        third = sympy.Rational(1, 3)
        two_thirds = sympy.Rational(2, 3)
        dofs: ListOfFunctionals = [
            PointEvaluation(reference, (two_thirds, third), entity=(1, 0)),
            PointEvaluation(reference, (third, two_thirds), entity=(1, 0)),
            PointEvaluation(reference, (0, third), entity=(1, 1)),
            PointEvaluation(reference, (0, two_thirds), entity=(1, 1)),
            PointEvaluation(reference, (sympy.Rational(1, 2), 0), entity=(1, 2)),
            PointEvaluation(reference, (third, third), entity=(2, 0))
        ]

        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_1d(reference.tdim, order)
        super().__init__(reference, order, poly, dofs, reference.tdim, 1)

    names = ["Fortin-Soulie", "FS"]
    references = ["triangle"]
    min_order = 2
    max_order = 2
    continuity = "L2"



"""Polynomial sets."""

import typing
from itertools import product

import sympy

from .functions import ScalarFunction, VectorFunction
from .symbols import AxisVariablesNotSingle, x


def polynomial_set_1d(
    dim: int, order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[ScalarFunction]:
    """One dimensional polynomial set.

    Args:
        dim: The number of variables
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    if dim == 1:
        return [ScalarFunction(variables[0] ** i) for i in range(order + 1)]
    if dim == 2:
        return [
            ScalarFunction(variables[0] ** i * variables[1] ** j)
            for j in range(order + 1)
            for i in range(order + 1 - j)
        ]
    if dim == 3:
        return [
            ScalarFunction(variables[0] ** i * variables[1] ** j * variables[2] ** k)
            for k in range(order + 1)
            for j in range(order + 1 - k)
            for i in range(order + 1 - k - j)
        ]
    raise ValueError(f"Unsupported dimension: {dim}")


def polynomial_set_vector(
    domain_dim: int, range_dim: int, order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[VectorFunction]:
    """Polynomial set.

    Args:
        domain_dim: The number of variables
        range_dim: The dimension of the output vector
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    set1d = polynomial_set_1d(domain_dim, order, variables)
    return [
        VectorFunction([p if i == j else 0 for j in range(range_dim)])
        for p in set1d
        for i in range(range_dim)
    ]


def Hdiv_polynomials(
    domain_dim: int, range_dim: int, order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[VectorFunction]:
    """Hdiv conforming polynomial set.

    Args:
        domain_dim: The number of variables
        range_dim: The dimension of the output vector
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    assert domain_dim == range_dim
    if domain_dim == 2:
        return [VectorFunction((
            variables[0] * variables[0] ** (order - 1 - j) * variables[1] ** j,
            variables[1] * variables[0] ** (order - 1 - j) * variables[1] ** j,
        )) for j in range(order)]
    if domain_dim == 3:
        basis: typing.List[VectorFunction] = []
        for j in range(order):
            for k in range(order - j):
                p = variables[0] ** (order - 1 - j - k) * variables[1] ** j * variables[2] ** k
                basis.append(VectorFunction((variables[0] * p, variables[1] * p, variables[2] * p)))
        return basis

    raise ValueError(f"Unsupported dimension: {domain_dim}")


def Hcurl_polynomials(
    domain_dim: int, range_dim: int, order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[VectorFunction]:
    """Hcurl conforming polynomial set.

    Args:
        domain_dim: The number of variables
        range_dim: The dimension of the output vector
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    assert domain_dim == range_dim
    if domain_dim == 2:
        return [VectorFunction((
            variables[0] ** (order - 1 - j) * variables[1] ** (j + 1),
            -variables[0] ** (order - j) * variables[1] ** j,
        )) for j in range(order)]
    if domain_dim == 3:
        poly: typing.List[VectorFunction] = []
        poly += [VectorFunction((
            variables[0] ** (m - 1) * variables[1] ** n * variables[2] ** (order - m - n + 1),
            0, -variables[0] ** m * variables[1] ** n * variables[2] ** (order - m - n)
        )) for n in range(order) for m in range(1, order + 1 - n)]
        poly += [VectorFunction((
            0, variables[0] ** m * variables[1] ** (n - 1) * variables[2] ** (order - m - n + 1),
            -variables[0] ** m * variables[1] ** n * variables[2] ** (order - m - n)
        )) for m in range(order) for n in range(1, order + 1 - m)]
        poly += [VectorFunction((
            variables[0] ** (order - n) * variables[1] ** n,
            -variables[0] ** (order + 1 - n) * variables[1] ** (n - 1), 0
        )) for n in range(1, order + 1)]
        return poly
    raise ValueError(f"Unsupported dimension: {domain_dim}")


def quolynomial_set_1d(
    dim: int, order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[ScalarFunction]:
    """One dimensional quolynomial set.

    Args:
        dim: The number of variables
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    basis = []
    for j in product(range(order + 1), repeat=dim):
        poly = ScalarFunction(1)
        for a, b in zip(variables, j):
            poly *= a ** b
        basis.append(poly)
    return basis


def quolynomial_set_vector(
    domain_dim: int, range_dim: int, order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[VectorFunction]:
    """Quolynomial set.

    Args:
        domain_dim: The number of variables
        range_dim: The dimension of the output vector
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    set1d = quolynomial_set_1d(domain_dim, order, variables)
    return [
        VectorFunction([p if i == j else 0 for j in range(range_dim)])
        for p in set1d
        for i in range(range_dim)
    ]


def Hdiv_quolynomials(
    domain_dim: int, range_dim: int, order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[VectorFunction]:
    """Hdiv conforming quolynomial set.

    Args:
        domain_dim: The number of variables
        range_dim: The dimension of the output vector
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    assert domain_dim == range_dim
    basis: typing.List[VectorFunction] = []
    for d in range(domain_dim):
        for j in product(range(order), repeat=domain_dim - 1):
            poly = 1
            for a, b in zip(variables, j[:d] + (order,) + j[d:]):
                poly *= a ** b
            basis.append(VectorFunction([poly if i == d else 0 for i in range(domain_dim)]))
    return basis


def Hcurl_quolynomials(
    domain_dim: int, range_dim: int, order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[VectorFunction]:
    """Hcurl conforming quolynomial set.

    Args:
        domain_dim: The number of variables
        range_dim: The dimension of the output vector
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    assert domain_dim == range_dim
    basis: typing.List[VectorFunction] = []
    for d in range(domain_dim):
        for j in product(
            *[range(order) if i == d else range(order + 1) for i in range(domain_dim)]
        ):
            if order not in j:
                continue
            poly = 1
            for a, b in zip(variables, j):
                poly *= a ** b
            basis.append(VectorFunction([poly if i == d else 0 for i in range(domain_dim)]))
    return basis


def serendipity_indices(
    total: int, linear: int, dim: int, done: typing.Optional[typing.List[int]] = None
) -> typing.List[typing.List[int]]:
    """Get the set indices for a serendipity polynomial set.

    Args:
        dim: The number of variables
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    if done is None:
        done = []
    if len(done) == dim:
        if done.count(1) >= linear:
            return [done]
        return []
    if len(done) == dim - 1:
        return serendipity_indices(total, linear, dim, done=done + [total - sum(done)])
    out = []
    for i in range(total - sum(done) + 1):
        out += serendipity_indices(total, linear, dim, done + [i])
    return out


def serendipity_set_1d(
    dim: int, order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[ScalarFunction]:
    """One dimensional serendipity set.

    Args:
        dim: The number of variables
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    basis: typing.List[ScalarFunction] = []
    for s in range(order + 1, order + dim + 1):
        for i in serendipity_indices(s, s - order, dim):
            p = 1
            for j, k in zip(variables, i):
                p *= j ** k
            basis.append(ScalarFunction(p))
    return basis


def serendipity_set_vector(
    domain_dim: int, range_dim: int, order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[VectorFunction]:
    """Serendipity set.

    Args:
        domain_dim: The number of variables
        range_dim: The dimension of the output vector
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    set1d = serendipity_set_1d(domain_dim, order, variables)
    return [
        VectorFunction([p if i == j else 0 for j in range(range_dim)])
        for p in set1d
        for i in range(range_dim)
    ]


def Hdiv_serendipity(
    domain_dim: int, range_dim: int, order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[VectorFunction]:
    """Hdiv conforming serendipity set.

    Args:
        domain_dim: The number of variables
        range_dim: The dimension of the output vector
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    assert domain_dim == range_dim
    if domain_dim == 2:
        return [
            VectorFunction((variables[0] ** (order + 1),
                            (order + 1) * variables[0] ** order * variables[1])),
            VectorFunction(((order + 1) * variables[0] * variables[1] ** order,
                            variables[1] ** (order + 1))),
        ]
    if domain_dim == 3:
        a = []
        if order == 0:
            a.append(VectorFunction((0, variables[0] * variables[2], -variables[0] * variables[1])))
            a.append(VectorFunction((variables[1] * variables[2], 0, -variables[0] * variables[1])))
        else:
            for i in range(order + 1):
                p = variables[1] ** i * variables[2] ** (order - i)
                a.append(VectorFunction((0, variables[0] * variables[2] * p,
                                         -variables[0] * variables[1] * p)))

                p = variables[0] ** i * variables[2] ** (order - i)
                a.append(VectorFunction((variables[1] * variables[2] * p, 0,
                                         -variables[0] * variables[1] * p)))

                p = variables[0] ** i * variables[1] ** (order - i)
                a.append(VectorFunction((variables[1] * variables[2] * p,
                                         -variables[0] * variables[2] * p, 0)))

        return [i.curl() for i in a]

    raise NotImplementedError()


def Hcurl_serendipity(
    domain_dim: int, range_dim: int, order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[VectorFunction]:
    """Hcurl conforming serendipity set.

    Args:
        domain_dim: The number of variables
        range_dim: The dimension of the output vector
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    assert domain_dim == range_dim
    if domain_dim == 2:
        return [
            VectorFunction(((order + 1) * variables[0] ** order * variables[1],
                            -variables[0] ** (order + 1))),
            VectorFunction((variables[1] ** (order + 1),
                            (order + 1) * variables[0] * variables[1] ** order)),
        ]
    if domain_dim == 3:
        out: typing.List[VectorFunction] = []
        if order == 1:
            out += [
                VectorFunction((0, variables[0] * variables[2], -variables[0] * variables[1])),
                VectorFunction((variables[1] * variables[2], 0, -variables[0] * variables[1])),
            ]
        else:
            for i in range(order):
                p = variables[0] ** i * variables[2] ** (order - 1 - i)
                out.append(VectorFunction((variables[1] * variables[2] * p, 0,
                                           -variables[0] * variables[1] * p)))

                p = variables[1] ** i * variables[2] ** (order - 1 - i)
                out.append(VectorFunction((0, variables[0] * variables[2] * p,
                                           -variables[0] * variables[1] * p)))

                p = variables[0] ** i * variables[1] ** (order - 1 - i)
                out.append(VectorFunction((variables[1] * variables[2] * p,
                                           -variables[0] * variables[2] * p, 0)))

        for p in serendipity_set_1d(domain_dim, order + 1):
            out.append(VectorFunction(tuple(p.diff(i) for i in variables)))
        return out

    raise NotImplementedError()


def prism_polynomial_set_1d(
    dim: int, order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[ScalarFunction]:
    """One dimensional polynomial set.

    Args:
        dim: The number of variables
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    assert dim == 3
    return [
        ScalarFunction(variables[0] ** i * variables[1] ** j * variables[2] ** k)
        for k in range(order + 1)
        for j in range(order + 1)
        for i in range(order + 1 - j)
    ]


def prism_polynomial_set_vector(
    domain_dim: int, range_dim: int, order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[VectorFunction]:
    """Polynomial set for a prism.

    Args:
        domain_dim: The number of variables
        range_dim: The dimension of the output vector
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    set1d = prism_polynomial_set_1d(domain_dim, order, variables)
    return [
        VectorFunction([p if i == j else 0 for j in range(range_dim)])
        for p in set1d
        for i in range(range_dim)
    ]


def pyramid_polynomial_set_1d(
    dim: int, order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[ScalarFunction]:
    """One dimensional polynomial set.

    Args:
        dim: The number of variables
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    assert dim == 3
    if order == 0:
        return [ScalarFunction(1)]

    poly = polynomial_set_1d(3, order)
    for d in range(order):
        for i in range(d + 1):
            for j in range(d + 1 - i):
                p = variables[0] ** i * variables[1] ** j * variables[2] ** (d - i - j)
                p *= (variables[0] * variables[1] / (1 - variables[2])) ** (order - d)
                poly.append(ScalarFunction(p))

    return poly


def pyramid_polynomial_set_vector(
    domain_dim: int, range_dim: int, order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[VectorFunction]:
    """Polynomial set for a pyramid.

    Args:
        domain_dim: The number of variables
        range_dim: The dimension of the output vector
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of polynomials
    """
    set1d = pyramid_polynomial_set_1d(domain_dim, order, variables)
    return [
        VectorFunction([p if i == j else 0 for j in range(range_dim)])
        for p in set1d
        for i in range(range_dim)
    ]


def _jrc(a, n) -> typing.Tuple[
    sympy.core.expr.Expr,
    sympy.core.expr.Expr,
    sympy.core.expr.Expr,
]:
    """Get the Jacobi recurrence relation coefficients.

    Args:
        a: The parameter a
        n: The parameter n

    Returns:
        The Jacobi coefficients
    """
    return (
        sympy.Rational((a + 2 * n + 1) * (a + 2 * n + 2), 2 * (n + 1) * (a + n + 1)),
        sympy.Rational(a * a * (a + 2 * n + 1), 2 * (n + 1) * (a + n + 1) * (a + 2 * n)),
        sympy.Rational(n * (a + n) * (a + 2 * n + 2), (n + 1) * (a + n + 1) * (a + 2 * n))
    )


def orthogonal_basis_interval(
    order: int, derivs: int, variables: AxisVariablesNotSingle = [x[0]]
) -> typing.List[typing.List[ScalarFunction]]:
    """Create a basis of orthogonal polynomials.

    Args:
        order: The maximum polynomial degree
        derivs: The number of derivatives to include
        variables: The variables to use

    Returns:
        A set of orthogonal polynomials
    """
    assert len(variables) == 1

    poly = [[ScalarFunction(1 if d == 0 else 0) for i in range(order + 1)]
            for d in range(derivs + 1)]
    for dx in range(derivs + 1):
        for i in range(1, order + 1):
            poly[dx][i] = poly[dx][i - 1] * (2 * variables[0] - 1) * (2 * i - 1) / i
            if dx > 0:
                poly[dx][i] += poly[dx - 1][i - 1] * 2 * dx * (2 * i - 1) / i
            if i > 1:
                poly[dx][i] -= (i - 1) * poly[dx][i - 2] / i
    return poly


def orthogonal_basis_triangle(
    order: int, derivs: int, variables: AxisVariablesNotSingle = [x[0], x[1]]
) -> typing.List[typing.List[ScalarFunction]]:
    """Create a basis of orthogonal polynomials.

    Args:
        order: The maximum polynomial degree
        derivs: The number of derivatives to include
        variables: The variables to use

    Returns:
        A set of orthogonal polynomials
    """
    assert len(variables) == 2

    def index(p: int, q: int) -> int:
        """Get the index."""
        return (p + q + 1) * (p + q) // 2 + q

    d_index = index

    poly = [[ScalarFunction(0) for i in range((order + 1) * (order + 2) // 2)]
            for d in range((derivs + 1) * (derivs + 2) // 2)]

    for dx in range(derivs + 1):
        for dy in range(derivs + 1 - dx):

            for p in range(order + 1):
                if p == 0:
                    poly[d_index(dx, dy)][index(0, p)] = ScalarFunction(1 if dx == dy == 0 else 0)
                else:
                    pinv = sympy.Rational(1, p)

                    poly[d_index(dx, dy)][index(0, p)] = (
                        poly[d_index(dx, dy)][index(0, p - 1)]
                        * (2 * variables[0] + variables[1] - 1) * (2 - pinv)
                    )
                    if dy > 0:
                        poly[d_index(dx, dy)][index(0, p)] += (
                            poly[d_index(dx, dy - 1)][index(0, p - 1)] * dy * (2 - pinv)
                        )
                    if dx > 0:
                        poly[d_index(dx, dy)][index(0, p)] += (
                            2 * dx * poly[d_index(dx - 1, dy)][index(0, p - 1)] * (2 - pinv)
                        )
                    if p > 1:
                        poly[d_index(dx, dy)][index(0, p)] -= (
                            poly[d_index(dx, dy)][index(0, p - 2)]
                            * (1 - variables[1]) ** 2 * (1 - pinv)
                        )
                        if dy > 0:
                            poly[d_index(dx, dy)][index(0, p)] += (
                                2 * dy * (1 - variables[1])
                                * poly[d_index(dx, dy - 1)][index(0, p - 2)] * (1 - pinv)
                            )
                        if dy > 1:
                            poly[d_index(dx, dy)][index(0, p)] -= (
                                dy * (dy - 1) * poly[d_index(dx, dy - 2)][index(0, p - 2)]
                                * (1 - pinv)
                            )

                for q in range(1, order - p + 1):
                    a, b, c = _jrc(2 * p + 1, q - 1)
                    poly[d_index(dx, dy)][index(q, p)] = (
                        poly[d_index(dx, dy)][index(q - 1, p)] * ((2 * variables[1] - 1) * a + b)
                    )
                    if q > 1:
                        poly[d_index(dx, dy)][index(q, p)] -= (
                            poly[d_index(dx, dy)][index(q - 2, p)] * c
                        )
                    if dy > 0:
                        poly[d_index(dx, dy)][index(q, p)] += (
                            poly[d_index(dx, dy - 1)][index(q - 1, p)] * 2 * dy * a
                        )
    return poly


def orthogonal_basis_quadrilateral(
    order: int, derivs: int, variables: AxisVariablesNotSingle = [x[0], x[1]]
) -> typing.List[typing.List[ScalarFunction]]:
    """Create a basis of orthogonal polynomials.

    Args:
        order: The maximum polynomial degree
        derivs: The number of derivatives to include
        variables: The variables to use

    Returns:
        A set of orthogonal polynomials
    """
    assert len(variables) == 2

    def d_index(p: int, q: int) -> int:
        """Get the derivative index."""
        return (p + q + 1) * (p + q) // 2 + q

    p0 = orthogonal_basis_interval(order, derivs, [variables[0]])
    p1 = orthogonal_basis_interval(order, derivs, [variables[1]])
    poly = [[ScalarFunction(0) for i in range((order + 1) ** 2)]
            for d in range((derivs + 1) * (derivs + 2) // 2)]
    for i in range(derivs + 1):
        for j in range(derivs + 1 - i):
            poly[d_index(i, j)] = [a * b for a in p0[i] for b in p1[j]]
    return poly


def orthogonal_basis_tetrahedron(
    order: int, derivs: int, variables: AxisVariablesNotSingle = x
) -> typing.List[typing.List[ScalarFunction]]:
    """Create a basis of orthogonal polynomials.

    Args:
        order: The maximum polynomial degree
        derivs: The number of derivatives to include
        variables: The variables to use

    Returns:
        A set of orthogonal polynomials
    """
    assert len(variables) == 3

    def index(p: int, q: int, r: int) -> int:
        """Get the index."""
        return (p + q + r) * (p + q + r + 1) * (p + q + r + 2) // 6 + (q + r) * (q + r + 1) // 2 + r

    d_index = index

    poly = [[ScalarFunction(0) for i in range((order + 1) * (order + 2) * (order + 3) // 6)]
            for d in range((derivs + 1) * (derivs + 2) * (derivs + 3) // 6)]
    for dx in range(derivs + 1):
        for dy in range(derivs + 1 - dx):
            for dz in range(derivs + 1 - dx - dy):

                for p in range(order + 1):
                    if p == 0:
                        poly[d_index(dx, dy, dz)][index(0, 0, p)] = ScalarFunction(
                            1 if dx == dy == dz == 0 else 0)
                    if p > 0:
                        invp = sympy.Rational(1, p)
                        poly[d_index(dx, dy, dz)][index(0, 0, p)] = (
                            poly[d_index(dx, dy, dz)][index(0, 0, p - 1)]
                            * (2 * variables[0] + variables[1] + variables[2] - 1) * (2 - invp)
                        )
                        if dx > 0:
                            poly[d_index(dx, dy, dz)][index(0, 0, p)] += (
                                poly[d_index(dx - 1, dy, dz)][index(0, 0, p - 1)]
                                * 2 * dx * (2 - invp)
                            )
                        if dy > 0:
                            poly[d_index(dx, dy, dz)][index(0, 0, p)] += (
                                poly[d_index(dx, dy - 1, dz)][index(0, 0, p - 1)] * dy * (2 - invp)
                            )
                        if dz > 0:
                            poly[d_index(dx, dy, dz)][index(0, 0, p)] += (
                                poly[d_index(dx, dy, dz - 1)][index(0, 0, p - 1)] * dz * (2 - invp)
                            )

                    if p > 1:
                        poly[d_index(dx, dy, dz)][index(0, 0, p)] -= (
                            poly[d_index(dx, dy, dz)][index(0, 0, p - 2)]
                            * (variables[1] + variables[2] - 1) ** 2 * (1 - invp)
                        )
                        if dy > 0:
                            poly[d_index(dx, dy, dz)][index(0, 0, p)] -= (
                                poly[d_index(dx, dy - 1, dz)][index(0, 0, p - 2)]
                                * 2 * (variables[1] + variables[2] - 1) * dy * (1 - invp)
                            )
                        if dy > 1:
                            poly[d_index(dx, dy, dz)][index(0, 0, p)] -= (
                                poly[d_index(dx, dy - 2, dz)][index(0, 0, p - 2)]
                                * dy * (dy - 1) * (1 - invp)
                            )
                        if dz > 0:
                            poly[d_index(dx, dy, dz)][index(0, 0, p)] -= (
                                poly[d_index(dx, dy, dz - 1)][index(0, 0, p - 2)]
                                * 2 * (variables[1] + variables[2] - 1) * dz * (1 - invp)
                            )
                        if dz > 1:
                            poly[d_index(dx, dy, dz)][index(0, 0, p)] -= (
                                poly[d_index(dx, dy, dz - 2)][index(0, 0, p - 2)]
                                * dz * (dz - 1) * (1 - invp)
                            )
                        if dz > 0 and dy > 0:
                            poly[d_index(dx, dy, dz)][index(0, 0, p)] -= (
                                poly[d_index(dx, dy - 1, dz - 1)][index(0, 0, p - 2)]
                                * 2 * dz * dy * (1 - invp)
                            )

                    for q in range(order - p + 1):
                        if q > 0:
                            a, b, c = _jrc(2 * p + 1, q - 1)
                            poly[d_index(dx, dy, dz)][index(0, q, p)] = (
                                poly[d_index(dx, dy, dz)][index(0, q - 1, p)]
                                * ((2 * variables[1] + variables[2] - 1) * a
                                   + (1 - variables[2]) * b)
                            )
                            if dy > 0:
                                poly[d_index(dx, dy, dz)][index(0, q, p)] += (
                                    poly[d_index(dx, dy - 1, dz)][index(0, q - 1, p)]
                                    * 2 * a * dy
                                )
                            if dz > 0:
                                poly[d_index(dx, dy, dz)][index(0, q, p)] += (
                                    poly[d_index(dx, dy, dz - 1)][index(0, q - 1, p)]
                                    * (a - b) * dz
                                )
                            if q > 1:
                                poly[d_index(dx, dy, dz)][index(0, q, p)] -= (
                                    poly[d_index(dx, dy, dz)][index(0, q - 2, p)]
                                    * (1 - variables[2]) ** 2 * c
                                )
                                if dz > 0:
                                    poly[d_index(dx, dy, dz)][index(0, q, p)] += (
                                        poly[d_index(dx, dy, dz - 1)][index(0, q - 2, p)]
                                        * 2 * (1 - variables[2]) * dz * c
                                    )
                                if dz > 1:
                                    poly[d_index(dx, dy, dz)][index(0, q, p)] -= (
                                        poly[d_index(dx, dy, dz - 2)][index(0, q - 2, p)]
                                        * dz * (dz - 1) * c
                                    )

                        for r in range(1, order - p - q + 1):
                            a, b, c = _jrc(2 * p + 2 * q + 2, r - 1)
                            poly[d_index(dx, dy, dz)][index(r, q, p)] = (
                                poly[d_index(dx, dy, dz)][index(r - 1, q, p)]
                                * ((variables[2] * 2 - 1) * a + b)
                            )
                            if dz > 0:
                                poly[d_index(dx, dy, dz)][index(r, q, p)] += (
                                    poly[d_index(dx, dy, dz - 1)][index(r - 1, q, p)] * 2 * a * dz
                                )
                            if r > 1:
                                poly[d_index(dx, dy, dz)][index(r, q, p)] -= (
                                    poly[d_index(dx, dy, dz)][index(r - 2, q, p)] * c
                                )

    return poly


def orthogonal_basis_hexahedron(
    order: int, derivs: int, variables: AxisVariablesNotSingle = x
) -> typing.List[typing.List[ScalarFunction]]:
    """Create a basis of orthogonal polynomials.

    Args:
        order: The maximum polynomial degree
        derivs: The number of derivatives to include
        variables: The variables to use

    Returns:
        A set of orthogonal polynomials
    """
    if variables is None:
        variables = x
    assert len(variables) == 3

    def d_index(p: int, q: int, r: int) -> int:
        """Get the derivative index."""
        return (p + q + r) * (p + q + r + 1) * (p + q + r + 2) // 6 + (q + r) * (q + r + 1) // 2 + r

    p0 = orthogonal_basis_interval(order, derivs, [variables[0]])
    p1 = orthogonal_basis_interval(order, derivs, [variables[1]])
    p2 = orthogonal_basis_interval(order, derivs, [variables[2]])
    poly = [[ScalarFunction(0) for i in range((order + 1) ** 3)]
            for d in range((derivs + 1) * (derivs + 2) * (derivs + 3) // 6)]
    for i in range(derivs + 1):
        for j in range(derivs + 1 - i):
            for k in range(derivs + 1 - i - j):
                poly[d_index(i, j, k)] = [a * b * c for a in p0[i] for b in p1[j] for c in p2[k]]
    return poly


def orthogonal_basis_prism(
    order: int, derivs: int, variables: AxisVariablesNotSingle = x
) -> typing.List[typing.List[ScalarFunction]]:
    """Create a basis of orthogonal polynomials.

    Args:
        order: The maximum polynomial degree
        derivs: The number of derivatives to include
        variables: The variables to use

    Returns:
        A set of orthogonal polynomials
    """
    if variables is None:
        variables = x
    assert len(variables) == 3

    def d_index_tri(p: int, q: int) -> int:
        """Get the derivative index for a triangle."""
        return (p + q + 1) * (p + q) // 2 + q

    def d_index(p: int, q: int, r: int) -> int:
        """Get the derivative index."""
        return (p + q + r) * (p + q + r + 1) * (p + q + r + 2) // 6 + (q + r) * (q + r + 1) // 2 + r

    p01 = orthogonal_basis_triangle(order, derivs, [variables[0], variables[1]])
    p2 = orthogonal_basis_interval(order, derivs, [variables[2]])
    poly = [[ScalarFunction(0) for i in range((order + 1) * (order + 1) * (order + 2) // 2)]
            for d in range((derivs + 1) * (derivs + 2) * (derivs + 3) // 6)]
    for i in range(derivs + 1):
        for j in range(derivs + 1 - i):
            for k in range(derivs + 1 - i - j):
                poly[d_index(i, j, k)] = [a * b for a in p01[d_index_tri(i, j)] for b in p2[k]]
    return poly


def orthogonal_basis_pyramid(
    order: int, derivs: int, variables: AxisVariablesNotSingle = x
) -> typing.List[typing.List[ScalarFunction]]:
    """Create a basis of orthogonal polynomials.

    Args:
        order: The maximum polynomial degree
        derivs: The number of derivatives to include
        variables: The variables to use

    Returns:
        A set of orthogonal polynomials
    """
    assert len(variables) == 3

    def index(i: int, j: int, k: int) -> int:
        """Get the index."""
        out = k + j * (order + 1) + i * (order + 1) * (order + 2) // 2 - i * (i ** 2 + 5) // 6
        if i > j:
            out -= i * (j - 1)
        else:
            out -= j * (j - 1) // 2 + i * (i - 1) // 2
        return out

    def d_index(p: int, q: int, r: int) -> int:
        """Get the derivative index."""
        return (p + q + r) * (p + q + r + 1) * (p + q + r + 2) // 6 + (q + r) * (q + r + 1) // 2 + r

    def combinations(n: int, k: int) -> int:
        """Get the number of combinations."""
        out = 1
        for i in range(n, n - k, -1):
            out *= i
        return out

    poly = [[ScalarFunction(0) for i in range((2 * order + 3) * (order + 2) * (order + 1) // 6)]
            for d in range((derivs + 1) * (derivs + 2) * (derivs + 3) // 6)]
    for dx in range(derivs + 1):
        for dy in range(derivs + 1 - dx):
            for dz in range(derivs + 1 - dx - dy):
                poly[d_index(dx, dy, dz)][0] = ScalarFunction(1 if dx == dy == dz == 0 else 0)

                for i in range(order + 1):
                    if i > 0:
                        poly[d_index(dx, dy, dz)][index(i, 0, 0)] = (
                            poly[d_index(dx, dy, dz)][index(i - 1, 0, 0)]
                            * (2 * variables[0] + variables[2] - 1) * (2 * i - 1) / i
                        )
                        if dx > 0:
                            poly[d_index(dx, dy, dz)][index(i, 0, 0)] += (
                                poly[d_index(dx - 1, dy, dz)][index(i - 1, 0, 0)]
                                * 2 * dx * (2 * i - 1) / i
                            )
                        if dz > 0:
                            poly[d_index(dx, dy, dz)][index(i, 0, 0)] += (
                                poly[d_index(dx, dy, dz - 1)][index(i - 1, 0, 0)]
                                * dz * (2 * i - 1) / i
                            )
                    if i > 1:
                        poly[d_index(dx, dy, dz)][index(i, 0, 0)] -= (
                            poly[d_index(dx, dy, dz)][index(i - 2, 0, 0)]
                            * (1 - variables[2]) ** 2 * (i - 1) / i
                        )
                        if dz > 0:
                            poly[d_index(dx, dy, dz)][index(i, 0, 0)] += (
                                poly[d_index(dx, dy, dz - 1)][index(i - 2, 0, 0)]
                                * 2 * (1 - variables[2]) * dz * (i - 1) / i
                            )
                        if dz > 1:
                            poly[d_index(dx, dy, dz)][index(i, 0, 0)] -= (
                                poly[d_index(dx, dy, dz - 2)][index(i - 2, 0, 0)]
                                * dz * (dz - 1) * (i - 1) / i
                            )

                    for j in range(order + 1):
                        if j > 0:
                            if i >= j:
                                poly[d_index(dx, dy, dz)][index(i, j, 0)] = (
                                    poly[d_index(dx, dy, dz)][index(i, j - 1, 0)]
                                    * (2 * variables[1] / (1 - variables[2]) - 1)
                                    * (2 * j - 1) / j
                                )
                                if dy > 0:
                                    poly[d_index(dx, dy, dz)][index(i, j, 0)] += (
                                        poly[d_index(dx, dy - 1, dz)][index(i, j - 1, 0)]
                                        * 2 * dy / (1 - variables[2])
                                        * (2 * j - 1) / j
                                    )
                                for di in range(1, dz + 1):
                                    poly[d_index(dx, dy, dz)][index(i, j, 0)] += (
                                        poly[d_index(dx, dy, dz - di)][index(i, j - 1, 0)]
                                        * combinations(dz, di)
                                        * variables[1] / (1 - variables[2]) ** (di + 1)
                                        * 2 * (2 * j - 1) / j
                                    )
                                    if dy > 0:
                                        poly[d_index(dx, dy, dz)][index(i, j, 0)] += (
                                            poly[d_index(dx, dy - 1, dz - di)][index(i, j - 1, 0)]
                                            * combinations(dz, di)
                                            * dy / (1 - variables[2]) ** (di + 1)
                                            * 2 * (2 * j - 1) / j
                                        )
                            else:
                                poly[d_index(dx, dy, dz)][index(i, j, 0)] = (
                                    poly[d_index(dx, dy, dz)][index(i, j - 1, 0)]
                                    * (2 * variables[1] + variables[2] - 1)
                                    * (2 * j - 1) / j
                                )
                                if dy > 0:
                                    poly[d_index(dx, dy, dz)][index(i, j, 0)] += (
                                        poly[d_index(dx, dy - 1, dz)][index(i, j - 1, 0)]
                                        * 2 * dy * (2 * j - 1) / j
                                    )
                                if dz > 0:
                                    poly[d_index(dx, dy, dz)][index(i, j, 0)] += (
                                        poly[d_index(dx, dy, dz - 1)][index(i, j - 1, 0)]
                                        * dz * (2 * j - 1) / j
                                    )
                        if j > 1:
                            if i >= j:
                                poly[d_index(dx, dy, dz)][index(i, j, 0)] -= (
                                    poly[d_index(dx, dy, dz)][index(i, j - 2, 0)]
                                    * (j - 1) / j
                                )
                            elif i + 1 == j:
                                poly[d_index(dx, dy, dz)][index(i, j, 0)] -= (
                                    poly[d_index(dx, dy, dz)][index(i, j - 2, 0)]
                                    * (1 - variables[2]) * (j - 1) / j
                                )
                                if dz > 0:
                                    poly[d_index(dx, dy, dz)][index(i, j, 0)] += (
                                        poly[d_index(dx, dy, dz - 1)][index(i, j - 2, 0)]
                                        * dz * (j - 1) / j
                                    )
                            else:
                                poly[d_index(dx, dy, dz)][index(i, j, 0)] -= (
                                    poly[d_index(dx, dy, dz)][index(i, j - 2, 0)]
                                    * (1 - variables[2]) ** 2 * (j - 1) / j
                                )
                                if dz > 0:
                                    poly[d_index(dx, dy, dz)][index(i, j, 0)] += (
                                        poly[d_index(dx, dy, dz - 1)][index(i, j - 2, 0)]
                                        * 2 * dz * (1 - variables[2]) * (j - 1) / j
                                    )
                                if dz > 1:
                                    poly[d_index(dx, dy, dz)][index(i, j, 0)] -= (
                                        poly[d_index(dx, dy, dz - 2)][index(i, j - 2, 0)]
                                        * dz * (dz - 1) * (j - 1) / j
                                    )

                        for k in range(1, order + 1 - max(i, j)):
                            a, b, c = _jrc(2 * max(i, j) + 2, k - 1)
                            poly[d_index(dx, dy, dz)][index(i, j, k)] = (
                                poly[d_index(dx, dy, dz)][index(i, j, k - 1)]
                                * a * (2 * variables[2] - 1)
                            )
                            if dz > 0:
                                poly[d_index(dx, dy, dz)][index(i, j, k)] += (
                                    poly[d_index(dx, dy, dz - 1)][index(i, j, k - 1)]
                                    * a * 2 * dz
                                )
                            poly[d_index(dx, dy, dz)][index(i, j, k)] += (
                                b * poly[d_index(dx, dy, dz)][index(i, j, k - 1)]
                            )
                            if k > 1:
                                poly[d_index(dx, dy, dz)][index(i, j, k)] -= (
                                    c * poly[d_index(dx, dy, dz)][index(i, j, k - 2)]
                                )

    return poly


def orthogonal_basis(
    cell: str, order: int, derivs: int, variables: typing.Optional[AxisVariablesNotSingle] = None
) -> typing.List[typing.List[ScalarFunction]]:
    """Create a basis of orthogonal polynomials.

    Args:
        cell: The cell type
        order: The maximum polynomial degree
        derivs: The number of derivatives to include
        variables: The variables to use

    Returns:
        A set of orthogonal polynomials
    """
    args: typing.List[typing.Any] = [order, derivs]
    if variables is not None:
        args.append(variables)

    if cell == "interval":
        return orthogonal_basis_interval(*args)
    if cell == "triangle":
        return orthogonal_basis_triangle(*args)
    if cell == "quadrilateral":
        return orthogonal_basis_quadrilateral(*args)
    if cell == "hexahedron":
        return orthogonal_basis_hexahedron(*args)
    if cell == "tetrahedron":
        return orthogonal_basis_tetrahedron(*args)
    if cell == "prism":
        return orthogonal_basis_prism(*args)
    if cell == "pyramid":
        return orthogonal_basis_pyramid(*args)

    raise ValueError(f"Unsupported cell type: {cell}")


def orthonormal_basis(
    cell: str, order: int, derivs: int, variables: typing.Optional[AxisVariablesNotSingle] = None
) -> typing.List[typing.List[ScalarFunction]]:
    """Create a basis of orthonormal polynomials.

    Args:
        cell: The cell type
        order: The maximum polynomial degree
        derivs: The number of derivatives to include
        variables: The variables to use

    Returns:
        A set of orthonormal polynomials
    """
    from .create import create_reference

    poly = orthogonal_basis(cell, order, derivs, variables)
    ref = create_reference(cell)
    if variables is None:
        variables = x
    norms = [sympy.sqrt((f ** 2).integral(ref, variables)) for f in poly[0]]
    for i, n in enumerate(norms):
        for j in range(len(poly)):
            poly[j][i] /= n
    return poly



"""Geometry."""

import typing

import sympy

PointType = typing.Tuple[sympy.core.expr.Expr, ...]
SetOfPoints = typing.Tuple[PointType, ...]
PointTypeInput = typing.Union[
    typing.Tuple[typing.Union[sympy.core.expr.Expr, int], ...],
    typing.List[typing.Union[sympy.core.expr.Expr, int]],
    sympy.matrices.dense.MutableDenseMatrix]
SetOfPointsInput = typing.Union[
    typing.Tuple[PointTypeInput, ...],
    typing.List[PointTypeInput]]


def _is_close(a: sympy.core.expr.Expr, b: int) -> bool:
    """Check if a Sympy expression is close to an int.

    Args:
        a: A Sympy expression
        b: An integer

    Returns:
        Is the sympy expression close to the integer?
    """
    return abs(a - b) < 1e-8


def parse_set_of_points_input(points: SetOfPointsInput) -> SetOfPoints:
    """Convert an input set of points to the correct format.

    Args:
        points: A set of points in some input format

    Returns:
        A set of points
    """
    return tuple(parse_point_input(p) for p in points)


def parse_point_input(point: PointTypeInput) -> PointType:
    """Convert an input point to the correct format.

    Args:
        point: A point in some input fotmat

    Returns:
        A point
    """
    if isinstance(point, sympy.Matrix):
        assert point.rows == 1 or point.cols == 1
        if point.rows == 1:
            return tuple(point[0, i] for i in range(point.cols))
        else:
            return tuple(point[i, 0] for i in range(point.rows))
    return tuple(sympy.S(i) for i in point)


def _vsub(v: PointType, w: PointType) -> PointType:
    """Subtract.

    Args:
        v: A vector
        w: A vector

    Returns:
        The vector v - w
    """
    return tuple(i - j for i, j in zip(v, w))


def _vdot(v: PointType, w: PointType) -> sympy.core.expr.Expr:
    """Compute dot product.

    Args:
        v: A vector
        w: A vector

    Returns:
        The dot product of v and w
    """
    out = sympy.Integer(0)
    for i, j in zip(v, w):
        out += i * j
    return out


def point_in_triangle(point: PointType, triangle: SetOfPoints) -> bool:
    """Check if a point is inside a triangle.

    Args:
        point: The point
        traingle: The vertices of the triangle

    Returns:
        Is the point inside the triangle?
    """
    v0 = _vsub(triangle[2], triangle[0])
    v1 = _vsub(triangle[1], triangle[0])
    v2 = _vsub(point, triangle[0])

    dot00 = _vdot(v0, v0)
    dot01 = _vdot(v0, v1)
    dot02 = _vdot(v0, v2)
    dot11 = _vdot(v1, v1)
    dot12 = _vdot(v1, v2)

    det = (dot00 * dot11 - dot01 * dot01)
    u = (dot11 * dot02 - dot01 * dot12) / det
    v = (dot00 * dot12 - dot01 * dot02) / det

    uv = u + v

    if isinstance(u, sympy.Float) and _is_close(u, 0):
        u = sympy.Integer(0)
    if isinstance(v, sympy.Float) and _is_close(v, 0):
        v = sympy.Integer(0)
    if isinstance(uv, sympy.Float) and _is_close(uv, 1):
        uv = sympy.Integer(1)

    return u >= 0 and v >= 0 and uv <= 1


def point_in_quadrilateral(point: PointType, quad: SetOfPoints) -> bool:
    """Check if a point is inside a quadrilateral.

    Args:
        point: The point
        traingle: The vertices of the quadrilateral

    Returns:
        Is the point inside the quadrilateral?
    """
    e0 = _vsub(quad[1], quad[0])
    e1 = _vsub(quad[0], quad[2])
    e2 = _vsub(quad[3], quad[1])
    e3 = _vsub(quad[2], quad[3])

    n0 = (-e0[1], e0[0])
    n1 = (-e1[1], e1[0])
    n2 = (-e2[1], e2[0])
    n3 = (-e3[1], e3[0])

    d0 = _vdot(n0, _vsub(point, quad[0]))
    d1 = _vdot(n1, _vsub(point, quad[2]))
    d2 = _vdot(n2, _vsub(point, quad[1]))
    d3 = _vdot(n3, _vsub(point, quad[3]))

    if isinstance(d0, sympy.Float) and _is_close(d0, 0):
        d0 = sympy.Integer(0)
    if isinstance(d1, sympy.Float) and _is_close(d1, 0):
        d1 = sympy.Integer(0)
    if isinstance(d2, sympy.Float) and _is_close(d2, 0):
        d2 = sympy.Integer(0)
    if isinstance(d3, sympy.Float) and _is_close(d3, 0):
        d3 = sympy.Integer(0)

    return d0 >= 0 and d1 >= 0 and d2 >= 0 and d3 >= 0


def point_in_tetrahedron(point: PointType, tetrahedron: SetOfPoints) -> bool:
    """Check if a point is inside a tetrahedron.

    Args:
        point: The point
        traingle: The vertices of the tetrahedron

    Returns:
        Is the point inside the tetrahedron?
    """
    v0 = _vsub(tetrahedron[3], tetrahedron[0])
    v1 = _vsub(tetrahedron[2], tetrahedron[0])
    v2 = _vsub(tetrahedron[1], tetrahedron[0])
    v3 = _vsub(point, tetrahedron[0])

    dot00 = _vdot(v0, v0)
    dot01 = _vdot(v0, v1)
    dot02 = _vdot(v0, v2)
    dot03 = _vdot(v0, v3)
    dot11 = _vdot(v1, v1)
    dot12 = _vdot(v1, v2)
    dot13 = _vdot(v1, v3)
    dot22 = _vdot(v2, v2)
    dot23 = _vdot(v2, v3)

    det = dot00 * (dot11 * dot22 - dot12 * dot12)
    det += dot01 * (dot02 * dot12 - dot01 * dot22)
    det += dot02 * (dot01 * dot12 - dot11 * dot02)

    u = (dot11 * dot22 - dot12 * dot12) * dot03
    u += (dot02 * dot12 - dot01 * dot22) * dot13
    u += (dot01 * dot12 - dot02 * dot11) * dot23
    u /= det
    v = (dot12 * dot02 - dot01 * dot22) * dot03
    v += (dot00 * dot22 - dot02 * dot02) * dot13
    v += (dot02 * dot01 - dot00 * dot12) * dot23
    v /= det
    w = (dot01 * dot12 - dot11 * dot02) * dot03
    w += (dot01 * dot02 - dot00 * dot12) * dot13
    w += (dot00 * dot11 - dot01 * dot01) * dot23
    w /= det

    uvw = u + v + w

    if isinstance(u, sympy.Float) and _is_close(u, 0):
        u = sympy.Integer(0)
    if isinstance(v, sympy.Float) and _is_close(v, 0):
        v = sympy.Integer(0)
    if isinstance(w, sympy.Float) and _is_close(w, 0):
        w = sympy.Integer(0)
    if isinstance(uvw, sympy.Float) and _is_close(uvw, 1):
        uvw = sympy.Integer(1)

    return u >= 0 and v >= 0 and w >= 0 and uvw <= 1



"""Serendipity elements on tensor product cells.

This element's definition appears in https://arxiv.org/abs/1809.02192
(Arbogast, Tao, 2018)
"""

from ..finite_element import DirectElement
from ..references import Reference
from ..symbols import x
from .dpc import DPC


class DirectSerendipity(DirectElement):
    """A direct serendipity element."""

    def __init__(self, reference: Reference, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
        """
        basis_functions = []
        basis_entities = []

        # Functions at vertices
        basis_functions += [a * b for a in (1 - x[1], x[1]) for b in (1 - x[0], x[0])]
        basis_entities += [(0, v) for v in range(4)]

        # Functions on edges
        if order >= 2:
            alpha_h = 1
            beta_h = 2
            gamma_h = 1
            xi_h = 2
            eta_h = 1

            alpha_v = 1
            beta_v = 2
            gamma_v = 1
            xi_v = 2
            eta_v = 1

            lambda_h = alpha_h * (1 - x[1]) + beta_h * x[1] + gamma_h
            r_h = -2 * x[1] / ((1 - x[1]) / xi_h + x[1] / eta_h)
            lambda_34 = xi_h * (1 - x[1]) + eta_h * x[1]

            lambda_v = alpha_v * (1 - x[0]) + beta_v * x[0] + gamma_v
            r_v = -2 * x[0] / ((1 - x[0]) / xi_v + x[0] / eta_v)
            lambda_12 = xi_v * (1 - x[0]) + eta_v * x[0]

            for j in range(order - 1):
                basis_functions.append((1 - x[1]) * x[1] * lambda_h ** j)
                basis_entities.append((1, 1))
            for j in range(order - 2):
                basis_functions.append((1 - x[1]) * x[1] * lambda_12 * lambda_h ** j)
                basis_entities.append((1, 2))
            basis_functions.append((1 - x[1]) * x[1] * r_v * lambda_h ** (order - 2))
            basis_entities.append((1, 2))

            for j in range(order - 1):
                basis_functions.append((1 - x[0]) * x[0] * lambda_v ** j)
                basis_entities.append((1, 0))
            for j in range(order - 2):
                basis_functions.append((1 - x[0]) * x[0] * lambda_34 * lambda_v ** j)
                basis_entities.append((1, 3))
            basis_functions.append((1 - x[0]) * x[0] * r_h * lambda_v ** (order - 2))
            basis_entities.append((1, 3))

        # Functions in interior
        if order >= 4:
            for f in DPC(reference, 4, "equispaced").get_basis_functions():
                basis_functions.append(f * x[0] * x[1] * (1 - x[0]) * (1 - x[1]))
                basis_entities.append((2, 0))

        super().__init__(reference, order, basis_functions, basis_entities,
                         reference.tdim, 1)

    names = ["direct serendipity"]
    references = ["quadrilateral"]
    min_order = 1
    continuity = "C0"



"""Test every element."""

import pytest

import symfem
from symfem import create_element
from symfem.functions import VectorFunction
from symfem.symbols import x
from symfem.utils import allequal

from .utils import test_elements


def test_all_tested():
    for e in symfem.create._elementlist:
        for r in e.references:
            if r == "dual polygon":
                continue
            for n in e.names:
                if n in test_elements[r]:
                    break
            else:
                raise ValueError(f"{e.names[0]} on a {r} is not tested")


@pytest.mark.parametrize("ref, element, order", [
    ("triangle", "Hermite", 4),
    ("tetrahedron", "Crouzeix-Raviart", 2)
])
def test_too_high_order(ref, element, order):
    with pytest.raises(ValueError):
        symfem.create_element(ref, element, order)


@pytest.mark.parametrize("ref, element, order", [
    ("triangle", "Hermite", 2),
    ("tetrahedron", "bubble", 3)
])
def test_too_low_order(ref, element, order):
    with pytest.raises(ValueError):
        symfem.create_element(ref, element, order)


@pytest.mark.parametrize(
    ("cell_type", "element_type", "order", "kwargs"),
    [[reference, element, order, kwargs]
     for reference, i in test_elements.items() for element, j in i.items()
     for kwargs, k in j for order in k])
def test_element(
    elements_to_test, cells_to_test, cell_type, element_type, order, kwargs,
    speed
):
    """Run tests for each element."""
    if elements_to_test != "ALL" and element_type not in elements_to_test:
        pytest.skip()
    if cells_to_test != "ALL" and cell_type not in cells_to_test:
        pytest.skip()
    if speed == "fast":
        if order > 2:
            pytest.skip()
        if order == 2 and cell_type in ["tetrahedron", "hexahedron", "prism", "pyramid"]:
            pytest.skip()

    space = create_element(cell_type, element_type, order, **kwargs)
    space.test()


@pytest.mark.parametrize("n_tri", [3, 4, 6, 8])
@pytest.mark.parametrize("order", range(2))
def test_dual_elements(elements_to_test, cells_to_test, n_tri, order):
    if elements_to_test != "ALL" and "dual" not in elements_to_test:
        pytest.skip()
    if cells_to_test != "ALL" and "dual polygon" not in cells_to_test:
        pytest.skip()

    space = create_element(f"dual polygon({n_tri})", "dual", order)
    sub_e = create_element("triangle", space.fine_space, space.order)
    for f, coeff_list in zip(space.get_basis_functions(), space.dual_coefficients):
        for piece, coeffs in zip(f.pieces.items(), coeff_list):
            fwd_map = VectorFunction(sub_e.reference.get_map_to(piece[0]))
            for dof, value in zip(sub_e.dofs, coeffs):
                point = fwd_map.subs(x, dof.point)
                assert allequal(value, piece[1].subs(x, point))


@pytest.mark.parametrize("n_tri", [3, 4])
@pytest.mark.parametrize("element_type", ["BC", "RBC"])
def test_bc_elements(elements_to_test, cells_to_test, n_tri, element_type):
    if elements_to_test != "ALL" and element_type not in elements_to_test:
        pytest.skip()
    if cells_to_test != "ALL" and "dual polygon" not in cells_to_test:
        pytest.skip()

    create_element(f"dual polygon({n_tri})", element_type, 1)



"""Test reordering inside tabulate_basis."""

import sympy

import symfem

half = sympy.Rational(1, 2)


def test_xxyyzz():
    element = symfem.create_element("triangle", "RT", 1)

    points = [(0, 1), (1, 0), (0, 0), (half, half)]
    r1 = element.tabulate_basis(points, "xyzxyz")
    r2 = element.tabulate_basis(points, "xxyyzz")
    for i, j in zip(r1, r2):
        assert i[0] == j[0]
        assert i[1] == j[3]
        assert i[2] == j[1]
        assert i[3] == j[4]
        assert i[4] == j[2]
        assert i[5] == j[5]


def test_vector():
    element = symfem.create_element("triangle", "RT", 1)

    points = [(0, 1), (1, 0), (0, 0), (half, half)]
    r1 = element.tabulate_basis(points, "xyzxyz")
    r2 = element.tabulate_basis(points, "xyz,xyz")
    for i, j in zip(r1, r2):
        assert i[0] == j[0][0]
        assert i[1] == j[0][1]
        assert i[2] == j[1][0]
        assert i[3] == j[1][1]
        assert i[4] == j[2][0]
        assert i[5] == j[2][1]



"""Values for Guzman-Neilan element."""

import sympy

coeffs = [
    [
        sympy.Rational(5, 8), sympy.Rational(5, 8), sympy.Rational(5, 8), 0, 0,
        sympy.Rational(125, 72), sympy.Rational(17, 36), sympy.Rational(17, 36),
        sympy.Rational(71, 18), sympy.Rational(-6343, 81000), sympy.Rational(-16157, 81000),
        sympy.Rational(325, 216), sympy.Rational(6391, 20250), sympy.Rational(742, 10125),
        sympy.Rational(58, 27), sympy.Rational(-7879, 10125), sympy.Rational(325, 216),
        sympy.Rational(-100093, 81000), sympy.Rational(-43907, 40500), sympy.Rational(58, 27),
        sympy.Rational(-20242, 10125), sympy.Rational(325, 216), sympy.Rational(-26609, 40500),
        sympy.Rational(-109907, 81000), sympy.Rational(58, 27), sympy.Rational(-34093, 40500),
        sympy.Rational(-45391, 20250), sympy.Rational(-13471, 54000), sympy.Rational(15971, 54000),
        sympy.Rational(-25, 27), sympy.Rational(-7879, 20250), sympy.Rational(-545, 144),
        sympy.Rational(10133, 40500), sympy.Rational(-545, 144), sympy.Rational(-26609, 81000),
        sympy.Rational(15359, 81000), sympy.Rational(39149, 13500), sympy.Rational(-25, 432),
        sympy.Rational(-9631, 3375), sympy.Rational(-6343, 162000), sympy.Rational(-16157, 162000),
        sympy.Rational(145, 36), sympy.Rational(-25, 432), sympy.Rational(63577, 27000),
        sympy.Rational(-62327, 27000),
    ],
    [
        sympy.Rational(15, 8), 0, 0, 0, sympy.Rational(125, 72), 0, sympy.Rational(1, 12),
        sympy.Rational(113, 36), sympy.Rational(-1, 3), sympy.Rational(-10921, 8100),
        sympy.Rational(27467, 16200), sympy.Rational(-25, 18), sympy.Rational(-21167, 8100),
        sympy.Rational(24767, 8100), sympy.Rational(-16, 9), sympy.Rational(25217, 16200),
        sympy.Rational(-25, 18), sympy.Rational(8533, 16200), sympy.Rational(6473, 2025),
        sympy.Rational(-16, 9), sympy.Rational(5833, 8100), sympy.Rational(125, 72),
        sympy.Rational(-8273, 4050), sympy.Rational(13967, 16200), sympy.Rational(44, 9),
        sympy.Rational(-8948, 2025), sympy.Rational(11267, 8100), sympy.Rational(9671, 1800),
        sympy.Rational(-2449, 450), sympy.Rational(25, 36), sympy.Rational(25217, 32400),
        sympy.Rational(235, 48), sympy.Rational(8533, 32400), sympy.Rational(-235, 144),
        sympy.Rational(-4967, 32400), sympy.Rational(13967, 32400), sympy.Rational(-9239, 1200),
        sympy.Rational(-25, 144), sympy.Rational(27467, 3600), sympy.Rational(-10921, 16200),
        sympy.Rational(6949, 4050), sympy.Rational(-35, 12), sympy.Rational(-145, 144),
        sympy.Rational(12967, 3600), sympy.Rational(-663, 400),
    ],
    [
        0, sympy.Rational(-15, 8), 0, 0, sympy.Rational(-61387, 81000),
        sympy.Rational(-39619, 40500), sympy.Rational(1, 3), sympy.Rational(-32381, 20250),
        sympy.Rational(-32869, 20250), sympy.Rational(-37, 18), sympy.Rational(59137, 81000),
        sympy.Rational(25, 18), sympy.Rational(-34, 9), sympy.Rational(27881, 20250),
        sympy.Rational(16, 9), sympy.Rational(2, 3), sympy.Rational(-125, 72),
        sympy.Rational(-9881, 40500), sympy.Rational(5, 3), sympy.Rational(-44, 9),
        sympy.Rational(-3131, 20250), sympy.Rational(25, 18), sympy.Rational(-13, 72),
        sympy.Rational(-1, 6), sympy.Rational(16, 9), sympy.Rational(-4, 9), 0,
        sympy.Rational(-3, 4), sympy.Rational(59, 72), sympy.Rational(-16631, 81000),
        sympy.Rational(1, 3), sympy.Rational(-8003, 4500), sympy.Rational(-11, 18),
        sympy.Rational(35, 12), sympy.Rational(-19003, 40500), sympy.Rational(-46369, 81000),
        sympy.Rational(-47, 9), sympy.Rational(14032, 10125), sympy.Rational(59, 18),
        sympy.Rational(-37, 36), sympy.Rational(-1, 72), sympy.Rational(-13369, 9000),
        sympy.Rational(-25, 36), sympy.Rational(1844, 375), sympy.Rational(-43631, 9000),
    ],
    [
        0, 0, sympy.Rational(15, 8), sympy.Rational(108281, 40500), 0,
        sympy.Rational(-75937, 81000), sympy.Rational(101531, 20250), sympy.Rational(-1, 3),
        sympy.Rational(-36281, 20250), sympy.Rational(-132187, 81000), sympy.Rational(-35, 72),
        sympy.Rational(125, 72), sympy.Rational(-145687, 40500), sympy.Rational(-47, 36),
        sympy.Rational(44, 9), sympy.Rational(25, 72), sympy.Rational(-25, 18), 0,
        sympy.Rational(13, 36), sympy.Rational(-16, 9), sympy.Rational(1, 12),
        sympy.Rational(-25, 18), sympy.Rational(15, 8), sympy.Rational(23203, 20250),
        sympy.Rational(-16, 9), sympy.Rational(41, 12), sympy.Rational(96187, 40500),
        sympy.Rational(-355, 144), sympy.Rational(635, 144), sympy.Rational(-21797, 40500),
        sympy.Rational(244687, 162000), sympy.Rational(-35, 12), sympy.Rational(-75937, 162000),
        sympy.Rational(82031, 9000), sympy.Rational(15, 16), sympy.Rational(5, 48),
        sympy.Rational(-64687, 18000), sympy.Rational(25, 36), sympy.Rational(63437, 18000),
        sympy.Rational(25, 48), sympy.Rational(-35, 144), sympy.Rational(-2194, 375),
        sympy.Rational(-52031, 81000), sympy.Rational(25, 16), sympy.Rational(-235, 144),
    ],
]



"""Test function classes."""

import sympy

from symfem.functions import MatrixFunction, ScalarFunction, VectorFunction
from symfem.piecewise_functions import PiecewiseFunction
from symfem.symbols import x


def test_scalar_function_add_sub():
    one = [1, ScalarFunction(1), sympy.Integer(1)]
    two = [2, ScalarFunction(2), sympy.Integer(2)]
    for i in one:
        for j in one:
            for k in two:
                assert i + j == k
                assert k - i == j


def test_scalar_function_mult_div():
    one = [1, ScalarFunction(1), sympy.Integer(1)]
    two = [2, ScalarFunction(2), sympy.Integer(2)]
    half = [ScalarFunction(sympy.Rational(1, 2)), sympy.Rational(1, 2)]
    for i in one:
        for j in two:
            for k in half:
                if not isinstance(i, int) or not isinstance(j, int):
                    assert i / j == k
                assert i == j * k


def test_scalar_function_neg():
    two = [2, ScalarFunction(2), sympy.Integer(2)]
    neg_two = [-2, ScalarFunction(-2), sympy.Integer(-2)]
    for i in two:
        for j in neg_two:
            if isinstance(i, ScalarFunction) or isinstance(j, ScalarFunction):
                assert -i == j


def test_scalar_function_subs():
    f = ScalarFunction(x[0] + x[1] ** 2)
    assert f.subs(x, [1, 0]) == 1
    assert f.subs(x, [1, 2]) == 5


def test_vector_function_add_sub():
    a = [(0, 1), VectorFunction((0, 1))]
    b = [(1, 1), VectorFunction((1, 1))]
    c = [(1, 2), VectorFunction((1, 2))]
    for i in a:
        for j in b:
            for k in c:
                if not isinstance(i, tuple) or not isinstance(j, tuple):
                    assert i + j == k
                if not isinstance(i, tuple) or not isinstance(k, tuple):
                    assert k - i == j


def test_vector_function_mul_div():
    two = [2, ScalarFunction(2), sympy.Integer(2)]
    a = [(0, 1), VectorFunction((0, 1))]
    a_double = [(0, 2), VectorFunction((0, 2))]
    for i in a:
        for j in two:
            for k in a_double:
                if not isinstance(i, tuple):
                    assert i * j == k
                    assert j * i == k
                if not isinstance(k, tuple):
                    assert k / j == i


def test_vector_function_neg():
    a = [(0, 1), VectorFunction((0, 1))]
    a_neg = [(0, -1), VectorFunction((0, -1))]
    for i in a:
        for j in a_neg:
            if isinstance(i, VectorFunction) or isinstance(j, VectorFunction):
                if not isinstance(i, tuple):
                    assert -i == j


def test_vector_function_subs():
    f = VectorFunction((x[0] + x[1] ** 2, x[1]))
    assert f.subs(x, [1, 0]) == (1, 0)
    assert f.subs(x, [1, 2]) == (5, 2)


def test_matrix_function_add_sub():
    m1 = [[0, 1], [-1, 0]]
    m2 = [[1, 1], [1, 1]]
    m3 = [[1, 2], [0, 1]]
    a = [m1, MatrixFunction(m1), sympy.Matrix(m1), MatrixFunction(sympy.Matrix(m1))]
    b = [m2, MatrixFunction(m2), sympy.Matrix(m2), MatrixFunction(sympy.Matrix(m2))]
    c = [m3, MatrixFunction(m3), sympy.Matrix(m3), MatrixFunction(sympy.Matrix(m3))]
    for i in a:
        for j in b:
            for k in c:
                # if isinstance(i, MatrixFunction) or isinstance(j, MatrixFunction):
                if isinstance(i, MatrixFunction):
                    assert i + j == k
                # if isinstance(i, MatrixFunction) or isinstance(k, MatrixFunction):
                if isinstance(k, MatrixFunction):
                    assert k - i == j


def test_matrix_function_mul_div():
    two = [2, ScalarFunction(2), sympy.Integer(2)]
    m3 = [[1, 2], [0, 1]]
    m4 = [[2, 4], [0, 2]]
    a = [m3, MatrixFunction(m3), sympy.Matrix(m3), MatrixFunction(sympy.Matrix(m3))]
    b = [m4, MatrixFunction(m4), sympy.Matrix(m4), MatrixFunction(sympy.Matrix(m4))]
    for i in a:
        for j in two:
            for k in b:
                if isinstance(i, MatrixFunction):
                    assert i * j == k
                    assert j * i == k
                if isinstance(k, MatrixFunction):
                    assert k / j == i


def test_matrix_function_neg():
    m3 = [[1, 2], [0, 1]]
    m3_neg = [[-1, -2], [0, -1]]
    a = [m3, MatrixFunction(m3), sympy.Matrix(m3), MatrixFunction(sympy.Matrix(m3))]
    a_neg = [m3_neg, MatrixFunction(m3_neg), sympy.Matrix(m3_neg),
             MatrixFunction(sympy.Matrix(m3_neg))]
    for i in a:
        for j in a_neg:
            if isinstance(i, MatrixFunction) or isinstance(j, MatrixFunction):
                if not isinstance(i, list):
                    assert -i == j


def test_matrix_function_subs():
    f = MatrixFunction(((x[0] + x[1] ** 2, x[1]), (x[0], 1)))
    assert f.subs(x, [1, 0]) == ((1, 0), (1, 1))
    assert f.subs(x, [1, 2]) == ((5, 2), (1, 1))


def test_piecewise_scalar_function_add_sub():
    f1 = PiecewiseFunction({
        ((0, 0), (1, 0), (0, 1)): x[0],
        ((1, 0), (1, 1), (0, 1)): x[1]
    }, 2)
    f2 = PiecewiseFunction({
        ((0, 0), (1, 0), (0, 1)): 0,
        ((1, 0), (1, 1), (0, 1)): x[0]
    }, 2)
    f3 = PiecewiseFunction({
        ((0, 0), (1, 0), (0, 1)): x[0],
        ((1, 0), (1, 1), (0, 1)): x[0] + x[1]
    }, 2)
    f4 = PiecewiseFunction({
        ((0, 0), (1, 0), (0, 1)): x[1],
        ((1, 0), (1, 1), (0, 1)): x[0] + x[1]
    }, 2)
    x1_pw = PiecewiseFunction({
        ((0, 0), (1, 0), (0, 1)): x[1],
        ((1, 0), (1, 1), (0, 1)): x[1]
    }, 2)

    assert f1 + f2 == f3
    assert f3 - f1 == f2

    for x1 in [x[1], ScalarFunction(x[1]), x1_pw]:
        assert f2 + x1 == f4
        assert x1 + f2 == f4

        assert f4 - x1 == f2


def test_piecewise_scalar_function_mult_div():
    f2 = PiecewiseFunction({
        ((0, 0), (1, 0), (0, 1)): 0,
        ((1, 0), (1, 1), (0, 1)): x[0]
    }, 2)
    f5 = PiecewiseFunction({
        ((0, 0), (1, 0), (0, 1)): 0,
        ((1, 0), (1, 1), (0, 1)): 2 * x[0]
    }, 2)

    two_pw = PiecewiseFunction({
        ((0, 0), (1, 0), (0, 1)): 2,
        ((1, 0), (1, 1), (0, 1)): 2
    }, 2)

    for two in [2, ScalarFunction(2), two_pw]:
        assert f2 * two == f5
        assert f5 / two == f2


def test_piecewise_scalar_function_neg():
    f2 = PiecewiseFunction({
        ((0, 0), (1, 0), (0, 1)): 0,
        ((1, 0), (1, 1), (0, 1)): x[0]
    }, 2)
    f6 = PiecewiseFunction({
        ((0, 0), (1, 0), (0, 1)): 0,
        ((1, 0), (1, 1), (0, 1)): -x[0]
    }, 2)

    assert f6 == -f2


def test_piecewise_scalar_function_subs():
    f2 = PiecewiseFunction({
        ((0, 0), (1, 0), (0, 1)): 0,
        ((1, 0), (1, 1), (0, 1)): x[0]
    }, 2)
    f7 = PiecewiseFunction({
        ((0, 0), (1, 0), (0, 1)): 0,
        ((1, 0), (1, 1), (0, 1)): 2
    }, 2)

    assert f2.subs(x[0], 2) == f7
    assert f2.subs(x[:2], (1, 1)) == 1



"""Functions to map functions between cells."""

from .functions import (AnyFunction, FunctionInput, MatrixFunction, VectorFunction,
                        parse_function_input)
from .geometry import PointType
from .symbols import x


def identity(
    f_in: FunctionInput, map: PointType, inverse_map: PointType, tdim: int
) -> AnyFunction:
    """Map functions.

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        tdim: The topological dimension of the cell

    Returns:
        The mapped function
    """
    return parse_function_input(f_in).subs(x, inverse_map)


def covariant(
    f_in: FunctionInput, map: PointType, inverse_map: PointType, tdim: int
) -> VectorFunction:
    """Map H(curl) functions.

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        tdim: The topological dimension of the cell

    Returns:
        The mapped function
    """
    f = parse_function_input(f_in).subs(x, inverse_map)
    assert f.is_vector

    j_inv = MatrixFunction([[i.diff(x[j]) for i in inverse_map] for j in range(tdim)])
    return j_inv @ f


def contravariant(
    f_in: FunctionInput, map: PointType, inverse_map: PointType, tdim: int
) -> VectorFunction:
    """Map H(div) functions.

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        tdim: The topological dimension of the cell

    Returns:
        The mapped function
    """
    f = parse_function_input(f_in).subs(x, inverse_map)
    assert f.is_vector

    jacobian = MatrixFunction([[i.diff(x[j]) for j in range(tdim)] for i in map])
    jacobian /= jacobian.det()
    return jacobian @ f


def double_covariant(
    f_in: FunctionInput, map: PointType, inverse_map: PointType, tdim: int
) -> MatrixFunction:
    """Map matrix functions.

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        tdim: The topological dimension of the cell

    Returns:
        The mapped function
    """
    f = parse_function_input(f_in).subs(x, inverse_map)
    assert f.is_matrix

    j_inv = MatrixFunction([[i.diff(x[j]) for i in inverse_map] for j in range(tdim)])
    return j_inv @ f @ j_inv.transpose()


def double_contravariant(
    f_in: FunctionInput, map: PointType, inverse_map: PointType, tdim: int
) -> MatrixFunction:
    """Map matrix functions.

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        tdim: The topological dimension of the cell

    Returns:
        The mapped function
    """
    f = parse_function_input(f_in).subs(x, inverse_map)
    assert f.is_matrix

    jacobian = MatrixFunction([[i.diff(x[j]) for j in range(tdim)] for i in map])
    jacobian /= jacobian.det()
    return jacobian @ f @ jacobian.transpose()



"""Test polynomials."""

from random import choice

import pytest
import sympy

from symfem import create_element, create_reference
from symfem.functions import VectorFunction
from symfem.polynomials import (Hcurl_polynomials, Hdiv_polynomials, orthogonal_basis,
                                orthonormal_basis)
from symfem.symbols import t, x


@pytest.mark.parametrize("reference", ["triangle", "tetrahedron"])
@pytest.mark.parametrize("order", range(1, 4))
def test_Hdiv_space(reference, order):
    ref = create_reference(reference)
    polynomials = Hdiv_polynomials(ref.tdim, ref.tdim, order)
    for p in polynomials:
        for i, j in zip(x, p):
            assert j / i == p[0] / x[0]


@pytest.mark.parametrize("reference", ["triangle", "tetrahedron"])
@pytest.mark.parametrize("order", range(1, 4))
def test_Hcurl_space(reference, order):
    ref = create_reference(reference)
    polynomials = Hcurl_polynomials(ref.tdim, ref.tdim, order)
    for p in polynomials:
        assert p.dot(VectorFunction(x[:ref.tdim])) == 0


@pytest.mark.parametrize("reference", ["triangle"])
def test_MTW_space(reference):
    e = create_element(reference, "MTW", 3)
    polynomials = e.get_polynomial_basis()
    for p in polynomials:
        assert p.div().as_sympy().is_real
        for vs in e.reference.sub_entities(1):
            sub_ref = create_reference(e.reference.sub_entity_types[1],
                                       [e.reference.vertices[i] for i in vs])
            p_edge = p.subs(x, [i + t[0] * j
                                for i, j in zip(sub_ref.origin, sub_ref.axes[0])])
            poly = p_edge.dot(sub_ref.normal()).as_sympy().expand().simplify()
            assert poly.is_real or sympy.Poly(poly).degree() <= 1


@pytest.mark.parametrize("reference", ["triangle", "quadrilateral",
                                       "tetrahedron", "hexahedron"])
@pytest.mark.parametrize("order", range(1, 5))
def test_BDFM_space(reference, order):
    e = create_element(reference, "BDFM", order)
    polynomials = e.get_polynomial_basis()
    tdim = e.reference.tdim
    for p in polynomials:
        for vs in e.reference.sub_entities(tdim - 1):
            sub_ref = create_reference(e.reference.sub_entity_types[tdim - 1],
                                       [e.reference.vertices[i] for i in vs])
            if tdim == 2:
                p_edge = p.subs(x, [i + t[0] * j
                                    for i, j in zip(sub_ref.origin, sub_ref.axes[0])])
            else:
                p_edge = p.subs(x, [i + t[0] * j + t[1] * k
                                    for i, j, k in zip(sub_ref.origin, sub_ref.axes[0],
                                                       sub_ref.axes[1])])
            poly = p_edge.dot(sub_ref.normal()).as_sympy().expand().simplify()
            assert poly.is_real or sympy.Poly(poly).degree() <= order - 1


@pytest.mark.parametrize("reference", [
    "interval", "triangle", "quadrilateral",
    "tetrahedron", "hexahedron", "prism",
    "pyramid"])
@pytest.mark.parametrize("order", range(5))
def test_orthogonal_polynomials(reference, order, speed):
    if speed == "fast" and order > 2:
        pytest.skip()

    polynomials = orthogonal_basis(reference, order, 0)[0]
    ref = create_reference(reference)
    if len(polynomials) <= 5:
        for i, p in enumerate(polynomials):
            for q in polynomials[:i]:
                assert (p * q).integral(ref, x) == 0
    else:
        for _ in range(15):
            p = choice(polynomials)
            q = choice(polynomials)
            if p != q:
                assert (p * q).integral(ref, x) == 0


@pytest.mark.parametrize("reference", [
    "interval",
    "triangle", "quadrilateral",
    "tetrahedron",
    "hexahedron", "prism",
    "pyramid"
])
@pytest.mark.parametrize("order", range(5))
def test_orthogonal_polynomial_derivatives(reference, order):
    polynomials = orthogonal_basis(reference, order, 2)
    ref = create_reference(reference)

    if ref.tdim == 1:
        first_d = [(1, 0)]
        second_d = [(2, 0, 0)]
    elif ref.tdim == 2:
        first_d = [(1, 0), (2, 1)]
        second_d = [(3, 0, 0), (4, 0, 1), (5, 1, 1)]
    elif ref.tdim == 3:
        first_d = [(1, 0), (2, 1), (3, 2)]
        second_d = [(4, 0, 0), (5, 0, 1), (6, 0, 2), (7, 1, 1), (8, 1, 2), (9, 2, 2)]
    else:
        raise NotImplementedError()

    for i, j in first_d:
        for p, q in zip(polynomials[0], polynomials[i]):
            assert p.diff(x[j]) == q
    for i, j, k in second_d:
        for p, q in zip(polynomials[0], polynomials[i]):
            assert p.diff(x[j]).diff(x[k]) == q


@pytest.mark.parametrize("reference", [
    "interval", "triangle", "quadrilateral",
    "tetrahedron", "hexahedron", "prism",
    "pyramid"])
@pytest.mark.parametrize("order", range(3))
def test_orthonormal_polynomials(reference, order, speed):
    if speed == "fast" and order > 2:
        pytest.skip()

    polynomials = orthonormal_basis(reference, order, 0)[0]
    ref = create_reference(reference)
    for p in polynomials:
        assert (p ** 2).integral(ref, x) == 1
